from myst.client import Client
from myst.openapi.models.backtest_create import BacktestCreate
from myst.openapi.models.backtest_get import BacktestGet


def request_sync(client: Client, project_uuid: str, json_body: BacktestCreate) -> BacktestGet:
    """Creates a backtest."""

    return client.request(
        method="post", path=f"/projects/{project_uuid}/backtests/", response_class=BacktestGet, request_model=json_body
    )



from typing import Union

from myst.models.timing import AbsoluteTiming, CronTiming, RelativeTiming, Timing
from myst.models.types import UNSET, OptionalArgument
from myst.openapi.models.absolute_timing_create import AbsoluteTimingCreate
from myst.openapi.models.cron_timing_create import CronTimingCreate
from myst.openapi.models.relative_timing_create import RelativeTimingCreate


def to_timing_create(
    timing: OptionalArgument[Timing],
) -> OptionalArgument[Union[AbsoluteTimingCreate, RelativeTimingCreate, CronTimingCreate]]:
    """Returns the appropriate OpenAPI model for the given timing, or None if timing is None."""
    if timing is None:
        return None
    elif timing is UNSET:
        return UNSET
    elif isinstance(timing, AbsoluteTiming):
        return AbsoluteTimingCreate(object="Timing", type="AbsoluteTiming", time=timing.time.to_iso_string())
    elif isinstance(timing, RelativeTiming):
        return RelativeTimingCreate(
            object="Timing",
            type="RelativeTiming",
            frequency=timing.frequency.to_iso_string() if timing.frequency else None,
            offset=timing.offset.to_iso_string() if timing.offset else None,
            time_zone=timing.time_zone,
        )
    elif isinstance(timing, CronTiming):
        return CronTimingCreate(
            object="Timing", type="CronTiming", cron_expression=timing.cron_expression, time_zone=timing.time_zone
        )
    else:
        raise TypeError("Argument `timing` must be of type `Timing` or None.")



from myst.client import Client
from myst.openapi.models.time_series_get import TimeSeriesGet
from myst.openapi.models.time_series_update import TimeSeriesUpdate


def request_sync(client: Client, project_uuid: str, uuid: str, json_body: TimeSeriesUpdate) -> TimeSeriesGet:
    """Updates a time series."""

    return client.request(
        method="patch",
        path=f"/projects/{project_uuid}/time_series/{uuid}",
        response_class=TimeSeriesGet,
        request_model=json_body,
    )



from abc import ABC, abstractmethod
from typing import Optional, Union

from myst.models.types import UUIDOrStr
from myst.resources.project import Project
from myst.resources.time_series import TimeSeries


class TimeSeriesRecipe(ABC):
    """A recipe from which a time series can be created."""

    @abstractmethod
    def create(
        self, project: Union["Project", UUIDOrStr], title: Optional[str] = None, description: Optional[str] = None
    ) -> TimeSeries:
        """Creates and returns a new time series from this recipe in the given project.

        Args:
            project: the project in which to create the time series
            title: the title of the time series; if none is given, the recipe will supply a sensible default
            description: the description of the time series

        Returns:
            the created time series
        """
        raise NotImplementedError()



from typing import Any, Dict

from myst.models import base_model


class SourceCreateParameters(base_model.BaseModel):
    """The parameters used in the node's connector. Must adhere to connector's parameter schema."""

    __root__: Dict[str, Any]

    def __getitem__(self, item: str) -> Any:
        return self.__root__[item]



from typing import List, Optional

from pydantic import Field
from typing_extensions import Literal

from myst.models import base_model
from myst.openapi.models.deploy_status import DeployStatus
from myst.openapi.models.operation_update_input_specs_schema import OperationUpdateInputSpecsSchema
from myst.openapi.models.operation_update_parameters import OperationUpdateParameters
from myst.openapi.models.time_dataset_spec import TimeDatasetSpec


class OperationUpdate(base_model.BaseModel):
    """Schema for operation update requests."""

    object_: Optional[Literal["Node"]] = Field(..., alias="object")
    uuid: Optional[str] = None
    create_time: Optional[str] = None
    update_time: Optional[str] = None
    organization: Optional[str] = None
    owner: Optional[str] = None
    type: Optional[Literal["Operation"]] = None
    title: Optional[str] = None
    description: Optional[str] = None
    project: Optional[str] = None
    deploy_status: Optional[DeployStatus] = None
    connector_uuid: Optional[str] = None
    parameters: Optional[OperationUpdateParameters] = None
    output_specs: Optional[List[TimeDatasetSpec]] = None
    input_specs_schema: Optional[OperationUpdateInputSpecsSchema] = None



from myst.client import Client
from myst.openapi.models.input_get import InputGet


def request_sync(client: Client, project_uuid: str, model_uuid: str, uuid: str) -> InputGet:
    """Deletes a model input."""

    return client.request(
        method="delete", path=f"/projects/{project_uuid}/models/{model_uuid}/inputs/{uuid}", response_class=InputGet
    )



from myst.client import Client
from myst.openapi.models.model_get import ModelGet
from myst.openapi.models.model_update import ModelUpdate


def request_sync(client: Client, project_uuid: str, uuid: str, json_body: ModelUpdate) -> ModelGet:
    """Updates a model."""

    return client.request(
        method="patch", path=f"/projects/{project_uuid}/models/{uuid}", response_class=ModelGet, request_model=json_body
    )



from typing import Optional, Union

from pydantic import Field
from typing_extensions import Literal

from myst.models import base_model
from myst.openapi.models.absolute_timing_create import AbsoluteTimingCreate
from myst.openapi.models.cron_timing_create import CronTimingCreate
from myst.openapi.models.hyperopt_create import HyperoptCreate
from myst.openapi.models.relative_timing_create import RelativeTimingCreate
from myst.openapi.models.search_space import SearchSpace


class HPOUpdate(base_model.BaseModel):
    """HPO schema for update input."""

    object_: Optional[Literal["HPO"]] = Field(..., alias="object")
    title: Optional[str] = None
    model: Optional[str] = None
    search_space: Optional[SearchSpace] = None
    test_start_time: Optional[str] = None
    test_end_time: Optional[str] = None
    fit_start_timing: Optional[Union[AbsoluteTimingCreate, RelativeTimingCreate]] = None
    fit_end_timing: Optional[Union[AbsoluteTimingCreate, RelativeTimingCreate]] = None
    fit_reference_timing: Optional[Union[AbsoluteTimingCreate, CronTimingCreate]] = None
    predict_start_timing: Optional[Union[AbsoluteTimingCreate, RelativeTimingCreate]] = None
    predict_end_timing: Optional[Union[AbsoluteTimingCreate, RelativeTimingCreate]] = None
    predict_reference_timing: Optional[CronTimingCreate] = None
    search_algorithm: Optional[HyperoptCreate] = None
    description: Optional[str] = None



from myst.client import Client
from myst.openapi.models.project_edges_list import ProjectEdgesList


def request_sync(client: Client, project_uuid: str) -> ProjectEdgesList:
    """List edges for a given project."""

    return client.request(method="get", path=f"/projects/{project_uuid}/edges/", response_class=ProjectEdgesList)



from myst.client import Client
from myst.openapi.models.time_series_insert import TimeSeriesInsert
from myst.openapi.models.time_series_insert_result_get import TimeSeriesInsertResultGet


def request_sync(
    client: Client, project_uuid: str, uuid: str, json_body: TimeSeriesInsert
) -> TimeSeriesInsertResultGet:
    """Manually insert time series data."""

    return client.request(
        method="post",
        path=f"/projects/{project_uuid}/time_series/{uuid}:insert",
        response_class=TimeSeriesInsertResultGet,
        request_model=json_body,
    )



from typing import List, Union

from myst.models import base_model


class TimeDatasetSpec(base_model.BaseModel):
    """A custom schema that represents the `spec` of a `TimeDataset`."""

    sample_period: str
    cell_shape: List[int]
    coordinate_labels: List[List[Union[int, str]]]
    axis_labels: List[Union[int, str]]



from myst.client import Client
from myst.openapi.models.source_get import SourceGet
from myst.openapi.models.source_update import SourceUpdate


def request_sync(client: Client, project_uuid: str, uuid: str, json_body: SourceUpdate) -> SourceGet:
    """Updates a source."""

    return client.request(
        method="patch",
        path=f"/projects/{project_uuid}/sources/{uuid}",
        response_class=SourceGet,
        request_model=json_body,
    )



from typing import List, Optional

from pydantic import Field
from typing_extensions import Literal

from myst.models import base_model
from myst.openapi.models.deploy_status import DeployStatus
from myst.openapi.models.model_get_input_specs_schema import ModelGetInputSpecsSchema
from myst.openapi.models.model_get_parameters import ModelGetParameters
from myst.openapi.models.time_dataset_spec import TimeDatasetSpec


class ModelGet(base_model.BaseModel):
    """Schema for model get responses."""

    object_: Literal["Node"] = Field(..., alias="object")
    uuid: str
    create_time: str
    organization: str
    owner: str
    type: Literal["Model"]
    title: str
    project: str
    creator: str
    deploy_status: DeployStatus
    connector_uuid: str
    parameters: ModelGetParameters
    input_specs_schema: ModelGetInputSpecsSchema
    update_time: Optional[str] = None
    description: Optional[str] = None
    output_specs: Optional[List[TimeDatasetSpec]] = None



from typing import Optional

from pydantic import Field
from typing_extensions import Literal

from myst.models import base_model
from myst.openapi.models.organization_role import OrganizationRole


class UserGet(base_model.BaseModel):
    """Schema for user get responses."""

    object_: Literal["User"] = Field(..., alias="object")
    uuid: str
    create_time: str
    email: str
    organization: str
    organization_role: OrganizationRole
    update_time: Optional[str] = None



from myst.client import Client
from myst.openapi.models.project_get import ProjectGet
from myst.openapi.models.project_update import ProjectUpdate


def request_sync(client: Client, uuid: str, json_body: ProjectUpdate) -> ProjectGet:
    """Updates a project."""

    return client.request(method="patch", path=f"/projects/{uuid}", response_class=ProjectGet, request_model=json_body)



from typing import List

from myst.models import base_model
from myst.openapi.models.model_run_result_list_item import ModelRunResultListItem


class ModelRunResultList(base_model.BaseModel):
    """Schema for model run result list responses."""

    data: List[ModelRunResultListItem]
    has_more: bool



from pydantic import Field
from typing_extensions import Literal

from myst.models import base_model


class OrganizationCreate(base_model.BaseModel):
    """Schema for organization create requests."""

    object_: Literal["Organization"] = Field(..., alias="object")
    name: str



from typing import Any, Dict, Iterable, List, Optional, Sequence

import numpy as np
import pandas as pd
from pydantic import Field, root_validator, validator
from pydantic.typing import Literal

from myst.core.data.time_array import TimeArray
from myst.core.data.validators import (
    check_arrays_same_length_as_time_range,
    ensure_not_none,
    maybe_to_boolean_int_array,
    to_numpy_float_array,
)
from myst.core.time.time import Time
from myst.core.time.time_delta import TimeDelta
from myst.models.base_model import BaseModel
from myst.models.types import AxisLabels, CoordinateLabels, Metadata, Shape


def _validate_time_array_compatibility(time_arrays: Iterable[TimeArray], check_shape: Optional[bool] = True) -> None:
    """Validates that the passed time arrays are compatible.

    Args:
        time_arrays: time arrays to validate
        check_shape: flag to indicate if we should check the time array shapes are equal

    Raises:
        ValueError: if the passed time arrays are incompatible
    """
    # Validate that the sample periods are the same across all time arrays.
    sample_period = None
    for time_array in time_arrays:
        if sample_period is not None:
            if time_array.sample_period != sample_period:
                raise ValueError("Time arrays do not share the same sample period.")
        else:
            sample_period = time_array.sample_period

    # Validate that the cell shapes are the same across all time arrays.
    cell_shape = None
    for time_array in time_arrays:
        if cell_shape is not None:
            if time_array.cell_shape != cell_shape:
                raise ValueError("Time arrays do not share the same cell shape.")
        else:
            cell_shape = time_array.cell_shape

    # Validate that the coordinate labels are the same across all time arrays.
    coordinate_labels = None
    for time_array in time_arrays:
        if coordinate_labels is not None:
            if not np.array_equal(time_array.coordinate_labels, coordinate_labels):
                raise ValueError("Time arrays do not share the same coordinate labels.")
        else:
            coordinate_labels = time_array.coordinate_labels

    # Validate that the axis labels are the same across all time arrays.
    axis_labels = None
    for time_array in time_arrays:
        if axis_labels is not None:
            if not np.array_equal(time_array.axis_labels, axis_labels):
                raise ValueError("Time arrays do not share the same axis labels.")
        else:
            axis_labels = time_array.axis_labels

    if check_shape:
        # Validate that the shapes are the same across all time arrays.
        shape = None
        for time_array in time_arrays:
            if shape is not None:
                if time_array.shape != shape:
                    raise ValueError("Time arrays do not share the same shape.")
            else:
                shape = time_array.shape


class TimeDatasetRow(BaseModel):
    start_time: Time = Field(
        ..., description="The time at which this time series data starts, inclusive.", example="2021-01-01T01:23:45Z"
    )
    end_time: Time = Field(
        ..., description="The time at which this time series data ends, exclusive.", example="2021-02-01T12:34:56Z"
    )
    as_of_time: Time = Field(
        ..., description="The time as of which this time series data was known.", example="2021-06-01T00:00:00Z"
    )
    values: np.ndarray = Field(
        ...,
        description="The multidimensional array of values for this time series.",
        examples=dict(
            empty=dict(summary="Empty array", value=[]),
            vector=dict(summary="1-D array of floats", value=[1.0, -47.47]),
            matrix=dict(summary="2-D array of floats", value=[[1.1, 1.2], [2.1, 2.2]]),
        ),
    )
    mask: Optional[np.ndarray] = Field(
        None,
        description=(
            "The mask to be used to filter out invalid or missing data. Must have the same shape as `values`, and each "
            "mask value must be either 0 or 1."
        ),
        examples=dict(
            empty=dict(summary="Empty array", value=[]),
            vector=dict(summary="1-D array of ints", value=[1, 0]),
            matrix=dict(summary="2-D array of ints", value=[[1, 0], [0, 1]]),
        ),
    )

    def dict(self, *args: Any, **kwargs: Any) -> Dict[str, Any]:
        """Returns the dictionary representation of this `TimeDatasetRow`."""
        dict_repr = super().dict(*args, **kwargs)

        arr: np.ndarray = dict_repr["values"]
        dict_repr["values"] = np.where(np.isnan(arr), None, arr).tolist()

        if self.mask is not None:
            dict_repr["mask"] = dict_repr["mask"].tolist()

        return dict_repr

    _ensure_values_not_none = validator("values", pre=True, allow_reuse=True)(ensure_not_none)
    _coerce_values_to_float_array = validator("values", pre=True, allow_reuse=True)(to_numpy_float_array)

    _coerce_mask_to_boolean_int_array = validator("mask", pre=True, allow_reuse=True)(maybe_to_boolean_int_array)

    class Config:
        # Allow arbitrary types in order to support `np.ndarray` directly.
        arbitrary_types_allowed = True


class TimeDataset(BaseModel):
    object: Literal["TimeDataset"] = Field(
        example="TimeDataset",
        description="The field representing the type of object this resource is. Must be `TimeDataset`.",
    )
    coordinate_labels: Optional[CoordinateLabels] = Field(
        example=("field",), description="The labels for each coordinate of a cell in the time dataset."
    )
    axis_labels: Optional[AxisLabels] = Field(
        example=(), description="The labels for each axis of a cell in the time dataset."
    )
    sample_period: TimeDelta = Field(
        example="PT1H", description="The sample period of the time dataset formatted as an ISO 8601 Duration string."
    )
    metadata: Optional[Metadata] = Field(
        None, example={"key": "value"}, description="User-defined metadata about the time dataset."
    )
    data: List[TimeDatasetRow] = Field(description="List of rows belonging to this time dataset.", min_items=1)
    cell_shape: Optional[Shape] = Field(example=(), description="The shape of each cell in the time dataset.")

    @validator("data")
    def ensure_unique_as_of_times(cls, v: List[TimeDatasetRow]) -> List[TimeDatasetRow]:
        as_of_times = [row.as_of_time for row in v]
        if len(set(as_of_times)) != len(as_of_times):
            raise ValueError("The `as_of_time` for each `TimeDatasetRow` must be unique.")

        return v

    @validator("data")
    def check_cell_shapes_equal(cls, v: List[TimeDatasetRow]) -> List[TimeDatasetRow]:
        if len(set([row.values.shape[1:] for row in v])) > 1:
            raise ValueError("The cell shape of `values` in each `TimeDatasetRow` must match.")

        return v

    @validator("data")
    def sort_rows_by_as_of_time_desc(cls, v: List[TimeDatasetRow]) -> List[TimeDatasetRow]:
        return sorted(v, key=lambda row: row.as_of_time, reverse=True)

    @validator("cell_shape", always=True)
    def infer_cell_shape_from_data(cls, v: Any, values: Dict[str, Any]) -> Any:
        data = values.get("data", None)

        # We rely on previous validation to ensure that each row has the same cell shape.
        if v is None and data is not None and len(data):
            return data[0].values.shape[1:]

        return v

    @root_validator
    def validate_time_range_consistent_with_values(cls, values: Dict[str, Any]) -> Dict[str, Any]:
        """Validates that the given time range is consistent with the values and mask."""
        sample_period = values.get("sample_period", None)

        if sample_period is not None:
            for row in values.get("data", ()):
                if row.start_time and row.end_time and row.values is not None:
                    check_arrays_same_length_as_time_range(
                        start_time=row.start_time,
                        end_time=row.end_time,
                        sample_period=sample_period,
                        values=row.values,
                        mask=row.mask,
                    )

        return values

    @classmethod
    def from_time_arrays(cls, time_arrays: Sequence[TimeArray]) -> "TimeDataset":
        if not time_arrays:
            raise ValueError("Cannot construct `TimeDataset` from empty collection.")

        _validate_time_array_compatibility(time_arrays=time_arrays, check_shape=False)

        first_time_array = time_arrays[0]

        time_dataset_rows = [
            TimeDatasetRow(
                start_time=time_array.start_time,
                end_time=time_array.end_time,
                as_of_time=time_array.as_of_time,
                values=time_array.values,
                mask=time_array.mask,
            )
            for time_array in time_arrays
        ]

        return TimeDataset(
            object="TimeDataset",
            cell_shape=first_time_array.cell_shape,
            sample_period=first_time_array.sample_period,
            data=time_dataset_rows,
            coordinate_labels=first_time_array.coordinate_labels,
            axis_labels=first_time_array.axis_labels,
        )

    def to_time_arrays(self) -> List[TimeArray]:
        """Returns the data in this time dataset as a list of time arrays."""
        return [
            TimeArray(
                sample_period=self.sample_period,
                start_time=time_dataset_row.start_time,
                end_time=time_dataset_row.end_time,
                as_of_time=time_dataset_row.as_of_time,
                values=time_dataset_row.values,
                mask=time_dataset_row.mask,
                coordinate_labels=self.coordinate_labels,
                axis_labels=self.axis_labels,
            )
            for time_dataset_row in self.data
        ]

    def flatten(self) -> TimeArray:
        """Flattens the time dataset by the latest as of time and returns a single time array.

        Deduplication happens by always prioritizing values of the latest as of time for each natural time.

        Note that missing values in the resulting time array will be masked using the `mask` attribute.

        Returns:
            flattened time array
        """
        time_arrays = self.to_time_arrays()

        # Special-case the common situation where the time dataset trivially reduces to a single time array.
        if len(time_arrays) == 1:
            return time_arrays[0]

        # Extract the pandas objects for the `values` and `mask` from the underlying time arrays.
        values_pandas_objects, mask_pandas_objects = {}, {}
        for time_array in self.to_time_arrays():
            values_pandas_object = time_array.to_pandas_object(apply_mask=False)
            values_pandas_objects.update({time_array.as_of_time.to_pandas_timestamp(): values_pandas_object})
            mask_pandas_object = values_pandas_object.copy()
            mask = (
                time_array.mask
                if time_array.mask is not None
                else np.zeros(shape=values_pandas_object.shape, dtype=int)
            )
            mask_pandas_object.values[:] = mask.reshape(mask_pandas_object.shape)
            mask_pandas_objects.update({time_array.as_of_time.to_pandas_timestamp(): mask_pandas_object})

        # Concatenate the extracted pandas objects across both dimensions.
        concat_values_pandas_object = pd.concat(values_pandas_objects, axis=0).sort_index()
        concat_mask_pandas_object = pd.concat(mask_pandas_objects, axis=0).sort_index()
        concat_combined_pandas_object = pd.concat([concat_values_pandas_object, concat_mask_pandas_object], axis=1)

        # Extract the time, as of time, and as of offset indexes.
        time_index = concat_values_pandas_object.index.get_level_values(level=1)
        as_of_time_index = concat_values_pandas_object.index.get_level_values(level=0)

        # Apply the boolean filter and deduplicate the time index based on the last row in each group.
        concat_combined_pandas_object = concat_combined_pandas_object.set_index(time_index)
        combined_pandas_object = concat_combined_pandas_object.groupby(concat_combined_pandas_object.index).last()

        # Ensure that the datetime index of the resulting pandas object is uniformly sampled.
        pandas_datetime_index = pd.date_range(
            start=combined_pandas_object.index.min(),
            end=combined_pandas_object.index.max() + self.sample_period.to_pandas_date_offset(),
            freq=self.sample_period.to_pandas_frequency(),
            name="time",
            closed="left",
            tz="UTC",
        )
        combined_pandas_object = combined_pandas_object.reindex(pandas_datetime_index)

        # Extract the pandas object and mask from the combined pandas object.
        num_values_columns = int(np.prod(self.cell_shape))
        values_pandas_object = combined_pandas_object.iloc[:, :num_values_columns]
        mask_pandas_object = combined_pandas_object.iloc[:, num_values_columns:].fillna(value=1)
        mask = mask_pandas_object.values.astype(int).reshape(-1, *self.cell_shape)

        # Extract the most recent as of time from the underlying time arrays.
        as_of_time = Time(as_of_time_index.max())

        if self.cell_shape == ():
            values_pandas_object = values_pandas_object.iloc[:, 0]

        return TimeArray.from_pandas_object(pandas_object=values_pandas_object, as_of_time=as_of_time, mask=mask)



import enum


@enum.unique
class DeployStatus(str, enum.Enum):
    NEW = "NEW"
    ACTIVE = "ACTIVE"
    INACTIVE = "INACTIVE"

    def __str__(self) -> str:
        return str(self.value)


@enum.unique
class OrgSharingRole(str, enum.Enum):
    EDITOR = "EDITOR"
    VIEWER = "VIEWER"

    def __str__(self) -> str:
        return str(self.value)



from myst.client import Client
from myst.openapi.models.hpo_job_get import HPOJobGet


def request_sync(client: Client, project_uuid: str, hpo_uuid: str, uuid: str) -> HPOJobGet:
    """Gets an hpo job for the project and specified hpo."""

    return client.request(
        method="get", path=f"/projects/{project_uuid}/hpos/{hpo_uuid}/jobs/{uuid}", response_class=HPOJobGet
    )



import json
from typing import TYPE_CHECKING, Any, Dict, Mapping, Optional, Union

import httpx
import pandas as pd

from myst.adapters.utils import get_resource_uuid
from myst.client import get_client
from myst.core.time.time import Time
from myst.models.types import UUIDOrStr
from myst.openapi.api.projects.backtests.results import get_backtest_result
from myst.resources.backtest_job import BacktestJob
from myst.resources.resource import Resource
from myst.resources.utils import backtest_result_to_pandas_data_frame

if TYPE_CHECKING:  # Avoid circular imports.
    from myst.resources.backtest import Backtest
    from myst.resources.project import Project


def _download_result(result_url: str) -> Dict[str, Any]:
    """Downloads the backtest result from the supplied URL and parses it."""
    response = httpx.get(result_url)

    if response.status_code == 200:
        result_data = json.loads(response.content)
    else:
        raise RuntimeError("Could not download backtest result.")

    return result_data


class BacktestResult(Resource):

    start_time: Time
    end_time: Time
    result_url: str
    metrics: Optional[Mapping[str, Optional[float]]]

    @classmethod
    def get(
        cls,
        project: Union["Project", UUIDOrStr],
        backtest: Union["Backtest", UUIDOrStr],
        job: Union[BacktestJob, UUIDOrStr],
    ) -> "BacktestResult":
        backtest_result = get_backtest_result.request_sync(
            client=get_client(),
            project_uuid=str(get_resource_uuid(project)),
            backtest_uuid=str(get_resource_uuid(backtest)),
            job_uuid=str(job.uuid) if isinstance(job, BacktestJob) else str(job),
        )

        return BacktestResult.parse_obj(backtest_result.dict())

    # TODO(CRYS-3649): move `backtest_result_to_pandas_data_frame` logic out of client.
    def to_pandas_data_frame(self) -> pd.DataFrame:
        """Downloads the backtest result and converts the time arrays to pandas data frames.

        Data will be re-indexed against the predictions' natural time index, dropping any target data that doesn't
        correspond to a prediction.

        Returns:
            a pandas data frame with the predictions made by the backtest, and their corresponding targets
        """
        result_data = _download_result(result_url=self.result_url)
        pandas_data_frame = backtest_result_to_pandas_data_frame(result_data=result_data)

        return pandas_data_frame



from typing import List

from myst.models import base_model
from myst.openapi.models.input_get import InputGet


class InputList(base_model.BaseModel):
    """Schema for input list responses."""

    data: List[InputGet]
    has_more: bool



from typing import List, Optional

from pydantic import Field
from typing_extensions import Literal

from myst.models import base_model
from myst.models.time_dataset import TimeDataset
from myst.openapi.models.model_run_result_get_inputs import ModelRunResultGetInputs


class ModelRunResultGet(base_model.BaseModel):
    """Schema for model run result get responses."""

    object_: Literal["NodeResult"] = Field(..., alias="object")
    uuid: str
    create_time: str
    type: Literal["ModelRunResult"]
    node: str
    start_time: str
    end_time: str
    as_of_time: str
    inputs: ModelRunResultGetInputs
    outputs: List[TimeDataset]
    update_time: Optional[str] = None



from typing import Optional

from pydantic import Field
from typing_extensions import Literal

from myst.models import base_model


class RelativeTimingGet(base_model.BaseModel):
    """Relative timing schema for get responses."""

    object_: Literal["Timing"] = Field(..., alias="object")
    type: Literal["RelativeTiming"]
    frequency: Optional[str] = None
    offset: Optional[str] = None
    time_zone: Optional[str] = None



from myst.client import Client
from myst.openapi.models.model_fit_result_list import ModelFitResultList


def request_sync(client: Client, project_uuid: str, model_uuid: str) -> ModelFitResultList:
    """Lists model fit results for the given model."""

    return client.request(
        method="get",
        path=f"/projects/{project_uuid}/models/{model_uuid}/fit_results/",
        response_class=ModelFitResultList,
    )



from typing import Any, Dict, List

from myst.models import base_model
from myst.models.time_dataset import TimeDataset


class ModelRunResultGetInputs(base_model.BaseModel):
    """The input data that was used to create the outputs of this result."""

    __root__: Dict[str, List[TimeDataset]]

    def __getitem__(self, item: str) -> Any:
        return self.__root__[item]



import datetime
import math
import re
from typing import Any, Generator, Optional, Tuple, Union

import pandas as pd
from pandas.tseries import offsets

PandasFrequency = Union[offsets.Day, offsets.Hour, offsets.Minute, offsets.Second]

_PANDAS_FREQUENCY_MAPPING = {
    offsets.Day: "days",
    offsets.Hour: "hours",
    offsets.Minute: "minutes",
    offsets.Second: "seconds",
}


# Define a regular expression to parse time delta ISO strings.
ISO_8601_PERIOD_REGEX = re.compile(
    r"^(?P<sign>[+-])?"
    r"P(?!\b)"
    r"(?P<years>[0-9]+([,.][0-9]+)?Y)?"
    r"(?P<months>[0-9]+([,.][0-9]+)?M)?"
    r"(?P<weeks>[0-9]+([,.][0-9]+)?W)?"
    r"(?P<days>[0-9]+([,.][0-9]+)?D)?"
    r"((?P<separator>T)(?P<hours>[0-9]+([,.][0-9]+)?H)?"
    r"(?P<minutes>[0-9]+([,.][0-9]+)?M)?"
    r"(?P<seconds>[0-9]+([,.][0-9]+)?S)?)?$"
)


def _parse_time_delta_string(iso_string: str) -> pd.DateOffset:
    regex_match = ISO_8601_PERIOD_REGEX.match(iso_string)

    if not regex_match:
        raise ValueError(f"Unable to parse ISO string `{iso_string}`.")

    regex_matched_groups = regex_match.groupdict()

    # Instantiate the `time_delta_dict`.
    time_delta_dict = dict()
    for time_delta_key, value in regex_matched_groups.items():
        if time_delta_key not in ("separator", "sign") and value is not None:

            # Parse the string time delta values.
            string_value = regex_matched_groups[time_delta_key][:-1].replace(",", ".")

            try:
                parsed_value = int(string_value)
            except ValueError:
                parsed_value = float(string_value)  # type: ignore

            if int(parsed_value) != parsed_value:
                raise ValueError(f"Passed value for `{time_delta_key}` must be an integer.")

            time_delta_dict[time_delta_key] = int(parsed_value)

    # If there is a negative sign at the beginning of the string, multiplier should be negative.
    if regex_matched_groups["sign"] == "-":
        time_delta_dict.update((key, value * -1) for key, value in time_delta_dict.items())

    return pd.DateOffset(**time_delta_dict)


def _format_time_delta(time_delta: pd.DateOffset) -> str:
    def update_sign(current_value: Union[int, float], current_sign: Optional[bool]) -> bool:
        if current_sign is None:
            if current_value >= 0:
                current_sign = True
            else:
                current_sign = False
        elif current_sign != (current_value >= 0):
            raise ValueError("Time delta cannot contain attributes that have mixed signs.")

        return current_sign

    # Set the default sign to undefined.
    sign = None

    # Format the date components of the time delta.
    formatted_time_delta = "P"
    for attribute, symbol in [("years", "Y"), ("months", "M"), ("weeks", "W"), ("days", "D")]:
        value = getattr(time_delta, attribute, None)

        if value is not None:
            if not isinstance(value, int):
                raise ValueError(f"Time delta `{attribute}` attribute must be an integer.")

            sign = update_sign(current_value=value, current_sign=sign)

            formatted_time_delta += f"{abs(value)}{symbol}"

    # Format the time components of the time delta.
    formatted_time_components = "T"
    for attribute, symbol in [("hours", "H"), ("minutes", "M")]:
        value = getattr(time_delta, attribute, None)

        if value is not None:
            sign = update_sign(current_value=value, current_sign=sign)

            formatted_time_components += f"{abs(value)}{symbol}"

    seconds = 0
    precision = 0
    for attribute, factor in [("seconds", 1), ("microseconds", 1e-6), ("nanoseconds", 1e-9)]:
        value = getattr(time_delta, attribute, None)

        if value is not None and value != 0:
            sign = update_sign(current_value=value, current_sign=sign)

            seconds += value * factor
            precision = abs(int(math.log10(factor)))

    if seconds != 0:
        formatted_time_components += f"{abs(seconds):.{precision}f}S"

    # Combine the components into the final formatted time delta.
    if len(formatted_time_components) > 1:
        formatted_time_delta += formatted_time_components

    # If needed, add the sign prefix to the formatted time delta.
    if not sign:
        formatted_time_delta = f"-{formatted_time_delta}"

    return formatted_time_delta


def _validate_time_delta(time_delta: pd.DateOffset) -> None:
    if not isinstance(time_delta, pd.DateOffset):
        raise TypeError("Time delta must be of type `pandas.DateOffset`.")

    for disallowed_attribute in [
        "year",
        "month",
        "day",
        "weekday",
        "hour",
        "minute",
        "second",
        "microsecond",
        "microseconds",
        "nanosecond",
        "nanoseconds",
    ]:
        if hasattr(time_delta, disallowed_attribute):
            raise ValueError(f"Time delta may not have `{disallowed_attribute}` attribute.")

    for allowed_attribute in ["years", "months", "weeks", "days", "hours", "minutes", "seconds"]:
        if hasattr(time_delta, allowed_attribute) and not isinstance(getattr(time_delta, allowed_attribute), int):
            raise ValueError(f"Time delta `{allowed_attribute}` attribute must be an integer.")


def _extract_pandas_date_offset_and_seconds(time_delta: "TimeDelta") -> Tuple[pd.DateOffset, float]:
    time_delta_pandas_date_offset = time_delta.to_pandas_date_offset()

    pandas_date_offset_kwargs = {}
    for attribute in ["years", "months", "weeks", "days"]:
        value = getattr(time_delta_pandas_date_offset, attribute, None)

        if value is not None and value != 0:
            pandas_date_offset_kwargs.update({attribute: value})

    pandas_date_offset = pd.DateOffset(**pandas_date_offset_kwargs)

    seconds = 0
    for attribute, factor in [("hours", 3600), ("minutes", 60), ("seconds", 1)]:
        value = getattr(time_delta_pandas_date_offset, attribute, None)

        if value is not None and value != 0:
            seconds += value * factor

    return pandas_date_offset, seconds


class TimeDelta:
    """Container for time delta objects that can be mapped from and to various third-party time objects.

    Note that `TimeDelta` respects calendar arithmetic for days, weeks, months, and years. This may cause mappings to
    some third-party time delta objects to fail, if they don't support this paradigm (e.g. `datetime.timedelta`).
    """

    def __init__(self, value: Union["TimeDelta", pd.DateOffset, PandasFrequency, datetime.timedelta, str]) -> None:
        """A `TimeDelta` container is initialized with a date offset, time delta, or ISO 8601 formatted string.

        The full ISO 8601 format specification is `PnYnMnDTnHnMnS`.

        Note that we only support integer values for all `n` and thus do not have support for sub-second precision.

        Args:
            value: date offset, time delta, or ISO 8601 formatted string
        """
        self._time_delta: pd.DateOffset

        if isinstance(value, TimeDelta):
            self._time_delta = value._time_delta
        elif isinstance(value, (offsets.Day, offsets.Hour, offsets.Minute, offsets.Second)):
            attribute = _PANDAS_FREQUENCY_MAPPING[type(value)]
            self._time_delta = pd.DateOffset(**{attribute: value.n})
        elif isinstance(value, pd.DateOffset):
            _validate_time_delta(value)
            self._time_delta = value
        elif isinstance(value, datetime.timedelta):
            if not value.total_seconds().is_integer():
                raise ValueError("Sub-second precision is not supported.")
            self._time_delta = pd.DateOffset(hours=value.days * 24, seconds=value.seconds)
        elif isinstance(value, str):
            self._time_delta = _parse_time_delta_string(value)
        else:
            raise TypeError(
                "Argument to `TimeDelta` constructor must be of type `TimeDelta`, `pd.DateOffset`, "
                "`datetime.timedelta`, or `str`."
            )

    def __eq__(self, other: object) -> bool:
        """Returns True if this time delta and `other` represent the same time delta, False otherwise."""
        return isinstance(other, TimeDelta) and _extract_pandas_date_offset_and_seconds(
            time_delta=self
        ) == _extract_pandas_date_offset_and_seconds(time_delta=other)

    def __ne__(self, other: object) -> bool:
        """Returns False if this time delta and `other` represent the same time delta, True otherwise."""
        return not self.__eq__(other)

    def __repr__(self) -> str:
        """Returns an encapsulated ISO 8601 formatted string representation for this time delta."""
        return f"<TimeDelta: {self.to_iso_string()}>"

    def __str__(self) -> str:
        """Returns the ISO 8601 formatted string representation for this time."""
        return self.to_iso_string()

    def to_iso_string(self) -> str:
        """Returns the ISO 8601 formatted string representation of this time."""
        return _format_time_delta(self._time_delta)

    def to_pandas_date_offset(self) -> pd.DateOffset:
        """Returns the `pandas.DateOffset` equivalent of this time delta."""
        return self._time_delta

    def to_pandas_frequency(self) -> pd.DateOffset:
        """Returns the pandas frequency equivalent of this time delta."""
        for attribute in ["years", "months", "weeks"]:
            if hasattr(self._time_delta, attribute):
                raise ValueError(f"Cannot convert `{attribute}` attribute, because calendar arithmetic is ambiguous.")

        pandas_frequency = None
        for attribute, pandas_date_offset_type in [
            ("days", offsets.Day),
            ("hours", offsets.Hour),
            ("minutes", offsets.Minute),
            ("seconds", offsets.Second),
        ]:
            value = getattr(self._time_delta, attribute, None)

            if value is not None:
                if pandas_frequency is not None:
                    raise ValueError("Cannot convert to frequency if multiple attributes are specified.")
                if value <= 0:
                    raise ValueError("Frequency must be positive.")

                pandas_frequency = pandas_date_offset_type(value)

        return pandas_frequency

    def to_pandas_timedelta(self) -> pd.Timedelta:
        """Returns the `pandas.Timedelta` equivalent of this time delta."""
        for attribute in ["years", "months", "weeks", "days"]:
            if hasattr(self._time_delta, attribute):
                raise ValueError(
                    f"Cannot convert `{attribute}` attribute, because calendar arithmetic is not supported."
                )

        pandas_timedelta_kwargs = {}
        for attribute in ["hours", "minutes", "seconds", "microseconds", "nanoseconds"]:
            value = getattr(self._time_delta, attribute, None)

            if value is not None:
                pandas_timedelta_kwargs[attribute] = value

        return pd.Timedelta(**pandas_timedelta_kwargs)

    def to_datetime_timedelta(self) -> datetime.timedelta:
        """Returns the `datetime.timedelta` equivalent of this time delta."""
        return self.to_pandas_timedelta().to_pytimedelta()

    @classmethod
    def __get_validators__(cls) -> Generator:
        yield cls.validate

    @classmethod
    def validate(cls, v: Any) -> "TimeDelta":
        return cls(v)



from typing import Any, Dict

from myst.client import Client
from myst.openapi.models.backtest_result_get import BacktestResultGet


def request_sync(client: Client, project_uuid: str, backtest_uuid: str, job_uuid: str) -> BacktestResultGet:
    """Gets a backtest result by its unique identifier."""

    params: Dict[str, Any] = {}
    params["job_uuid"] = job_uuid

    params = {k: v for k, v in params.items() if v is not None}

    return client.request(
        method="get",
        path=f"/projects/{project_uuid}/backtests/{backtest_uuid}/results/",
        response_class=BacktestResultGet,
        params=params,
    )



from myst.client import Client
from myst.openapi.models.time_series_run_policy_get import TimeSeriesRunPolicyGet


def request_sync(client: Client, project_uuid: str, time_series_uuid: str, uuid: str) -> TimeSeriesRunPolicyGet:
    """Deletes a time series run policy."""

    return client.request(
        method="delete",
        path=f"/projects/{project_uuid}/time_series/{time_series_uuid}/run_policies/{uuid}",
        response_class=TimeSeriesRunPolicyGet,
    )



from typing import Any, List, Optional

from pydantic import Field
from typing_extensions import Literal

from myst.models import base_model


class TimeSeriesCreate(base_model.BaseModel):
    """Schema for time series create requests."""

    title: str
    sample_period: str
    cell_shape: List[Any]
    coordinate_labels: List[Any]
    axis_labels: List[Any]
    object_: Optional[Literal["Node"]] = Field("Node", alias="object")
    type: Optional[Literal["TimeSeries"]] = "TimeSeries"
    description: Optional[str] = None



from typing import Any, List, Optional, Union

from pydantic import Field
from typing_extensions import Literal

from myst.models import base_model


class InputUpdate(base_model.BaseModel):
    """Schema for input update requests."""

    object_: Optional[Literal["Edge"]] = Field(..., alias="object")
    uuid: Optional[str] = None
    create_time: Optional[str] = None
    update_time: Optional[str] = None
    type: Optional[Literal["Input"]] = None
    downstream_node: Optional[str] = None
    upstream_node: Optional[str] = None
    output_index: Optional[int] = None
    label_indexer: Optional[Union[List[Union[int, str, List[Any]]], List[Any], int, str]] = None
    group_name: Optional[str] = None



from myst.models import base_model
from myst.models.time_dataset import TimeDataset


class TimeSeriesInsert(base_model.BaseModel):
    """Schema for time series insert requests."""

    time_dataset: TimeDataset



from typing import Any, Dict, Union

from myst.models import base_model


class TimeDatasetMetadata(base_model.BaseModel):
    """User-defined metadata about the time dataset."""

    __root__: Dict[str, Union[bool, float, int, str]]

    def __getitem__(self, item: str) -> Any:
        return self.__root__[item]



from typing import Optional, Union

import pydantic

from myst.core.time.time import Time
from myst.core.time.time_delta import TimeDelta
from myst.models.base_model import BaseModel
from myst.models.types import OptionalArgument, Unset


class Timing(BaseModel):
    class Config:
        extra = pydantic.Extra.ignore


class AbsoluteTiming(Timing):
    time: Time


class RelativeTiming(Timing):
    frequency: Optional[TimeDelta] = None
    offset: Optional[TimeDelta] = None
    time_zone: Optional[str] = "UTC"


class CronTiming(Timing):
    cron_expression: str
    time_zone: Optional[str] = "UTC"


ScheduleTiming = Union[AbsoluteTiming, CronTiming, RelativeTiming]
AbsoluteOrRelativeTiming = Union[AbsoluteTiming, RelativeTiming]
AbsoluteOrCronTiming = Union[AbsoluteTiming, CronTiming]
TimeRangeBoundary = Union[Time, TimeDelta, AbsoluteOrRelativeTiming]


def from_time_range_boundary(
    timing: OptionalArgument[Union[Time, TimeDelta, AbsoluteOrRelativeTiming]]
) -> OptionalArgument[AbsoluteOrRelativeTiming]:
    """Returns the given argument as a timing appropriate for a time range boundary.

    If an `AbsoluteOrRelativeTiming`, `None` or `UNSET` is passed, it will be returned unchanged.

    If a `Time` is passed, it will be interpreted as specifying a fixed point in time, and an `AbsoluteTiming` will be
    returned.

    If a `TimeDelta` is passed, it will be interpreted as the "offset" of a `RelativeTiming`. The corresponding
    `RelativeTiming` will be returned.

    Args:
        timing: the input time specifier to be interpreted

    Returns:
        the timing, after interpretation

    Raises:
        TypeError: if the argument is of a type that cannot be understood as a time range boundary
    """
    if timing is None or isinstance(timing, (AbsoluteTiming, RelativeTiming)):
        return timing
    elif isinstance(timing, Unset):
        return timing
    elif isinstance(timing, Time):
        # Interpret as fixed point in time.
        return AbsoluteTiming(time=timing)
    elif isinstance(timing, TimeDelta):
        # Interpret as offset (notably, not frequency).
        return RelativeTiming(offset=timing)
    else:
        raise TypeError("Argument must be of type `Time`, `TimeDelta, `Timing`, or None.")


def from_schedule_specifier(
    timing: OptionalArgument[Union[Time, TimeDelta, ScheduleTiming]]
) -> OptionalArgument[Optional[ScheduleTiming]]:
    """Returns the given argument as a timing appropriate for scheduling.

    If a `None` is passed, `None` will be returned.

    If a `UNSET` is passed, `UNSET` will be returned.

    If a `Time` is passed, it will be interpreted as specifying a fixed point in time, and an `AbsoluteTiming` will
    be returned.

    If a `TimeDelta` is passed, it will be interpreted as the "frequency" of a `RelativeTiming`. The corresponding
    `RelativeTiming` will be returned.

    If a `ScheduleTiming` is passed, it will be returned unaltered.

    Args:
        timing: the input time specifier to be interpreted

    Returns:
        the timing, after interpretation

    Raises:
        TypeError: if the argument is of a type that cannot be understood as a schedule timing
    """
    if timing is None:
        return timing
    elif isinstance(timing, Unset):
        return timing
    elif isinstance(timing, Time):
        return AbsoluteTiming(time=timing)
    elif isinstance(timing, TimeDelta):
        return RelativeTiming(frequency=timing)
    elif isinstance(timing, (AbsoluteTiming, RelativeTiming, CronTiming)):
        return timing
    else:
        raise TypeError("Argument must be of type `Time`, `TimeDelta, or `ScheduleTiming`.")



from typing import List
from uuid import UUID

from myst.connectors.operation_connector import OperationConnector


class NumericalExpression(OperationConnector):
    def __init__(self, variable_names: List[str], math_expression: str) -> None:
        super().__init__(
            uuid=UUID("931acb25-7ace-47f9-9526-1ba8ddf06f14"),
            parameters=dict(variable_names=variable_names, math_expression=math_expression),
        )



from typing import TYPE_CHECKING, Any, Dict, Union
from uuid import UUID

from myst.models.types import UNSET

if TYPE_CHECKING:  # Avoid circular imports.
    from myst.models.types import UUIDOrStr
    from myst.resources.resource import Resource


def drop_unset_values(input_dict: Dict[str, Any]) -> Dict[str, Any]:
    """Drops `UNSET` values from the specified dictionary.

    Args:
        input_dict: the dictionary to drop `UNSET` values from

    Returns:
        the resulting dictionary
    """
    output_dict = {key: value for key, value in input_dict.items() if value is not UNSET}

    return output_dict


def get_resource_uuid(resource: Union["Resource", "UUIDOrStr"]) -> UUID:
    """Extracts the UUID from the specified resource."""
    if isinstance(resource, UUID):
        return resource
    if isinstance(resource, str):
        return UUID(resource)
    else:
        return resource.uuid



from typing import Optional, Union

from pydantic import Field
from typing_extensions import Literal

from myst.models import base_model
from myst.openapi.models.absolute_timing_get import AbsoluteTimingGet
from myst.openapi.models.cron_timing_get import CronTimingGet
from myst.openapi.models.relative_timing_get import RelativeTimingGet


class ModelFitPolicyGet(base_model.BaseModel):
    """Schema for model fit policy get responses."""

    object_: Literal["Policy"] = Field(..., alias="object")
    uuid: str
    create_time: str
    type: Literal["ModelFitPolicy"]
    creator: str
    schedule_timing: Union[AbsoluteTimingGet, RelativeTimingGet, CronTimingGet]
    active: bool
    node: str
    start_timing: Union[AbsoluteTimingGet, RelativeTimingGet]
    end_timing: Union[AbsoluteTimingGet, RelativeTimingGet]
    update_time: Optional[str] = None



from myst.client import Client
from myst.openapi.models.time_series_get import TimeSeriesGet


def request_sync(client: Client, uuid: str) -> TimeSeriesGet:
    """Gets a time series by its unique identifier."""

    return client.request(method="get", path=f"/projects/-/time_series/{uuid}", response_class=TimeSeriesGet)



from typing import Any, Dict, Optional

from myst.models import base_model


class BacktestResultGetMetrics(base_model.BaseModel):
    """"""

    __root__: Dict[str, Optional[float]]  # manual intervention required

    def __getitem__(self, item: str) -> Any:
        return self.__root__[item]



from typing import List, Optional

from pydantic import Field
from typing_extensions import Literal

from myst.models import base_model
from myst.openapi.models.job_error import JobError
from myst.openapi.models.job_state import JobState


class ModelFitJobGet(base_model.BaseModel):
    """Model fit job schema for get responses."""

    object_: Literal["NodeJob"] = Field(..., alias="object")
    uuid: str
    create_time: str
    type: Literal["ModelFitJob"]
    node: str
    errors: List[JobError]
    start_time: str
    end_time: str
    as_of_time: str
    state: JobState
    update_time: Optional[str] = None
    result: Optional[str] = None
    error_reason: Optional[str] = None
    error_edge: Optional[str] = None



from myst.client import Client
from myst.openapi.models.time_series_run_result_get import TimeSeriesRunResultGet


def request_sync(client: Client, project_uuid: str, time_series_uuid: str, uuid: str) -> TimeSeriesRunResultGet:
    """Gets a time series run result by its unique identifier."""

    return client.request(
        method="get",
        path=f"/projects/{project_uuid}/time_series/{time_series_uuid}/run_results/{uuid}",
        response_class=TimeSeriesRunResultGet,
    )



from myst.client import Client
from myst.openapi.models.deployment_get import DeploymentGet


def request_sync(client: Client, project_uuid: str, uuid: str) -> DeploymentGet:
    """Deactivates a deployment for a given project."""

    return client.request(
        method="post", path=f"/projects/{project_uuid}/deployments/{uuid}:deactivate", response_class=DeploymentGet
    )



from typing import Optional

from pydantic import Field
from typing_extensions import Literal

from myst.models import base_model
from myst.openapi.models.model_create_parameters import ModelCreateParameters


class ModelCreate(base_model.BaseModel):
    """Schema for model create requests."""

    title: str
    connector_uuid: str
    object_: Optional[Literal["Node"]] = Field("Node", alias="object")
    type: Optional[Literal["Model"]] = "Model"
    description: Optional[str] = None
    parameters: Optional[ModelCreateParameters] = None



from myst.client import Client
from myst.openapi.models.backtest_get import BacktestGet


def request_sync(client: Client, project_uuid: str, uuid: str) -> BacktestGet:
    """Copies a backtest."""

    return client.request(
        method="post", path=f"/projects/{project_uuid}/backtests/{uuid}:copy", response_class=BacktestGet
    )



from typing import Optional

from pydantic import Field
from typing_extensions import Literal

from myst.models import base_model
from myst.openapi.models.job_state import JobState


class BacktestJobGet(base_model.BaseModel):
    """Abstract base job schema for get responses."""

    object_: Literal["BacktestJob"] = Field(..., alias="object")
    uuid: str
    backtest: str
    state: JobState
    seconds_to_complete: Optional[int] = None



import warnings
from typing import Any, Dict

import pandas as pd

from myst.models.time_dataset import TimeDataset


def backtest_result_to_pandas_data_frame(result_data: Dict[str, Any]) -> pd.DataFrame:
    """Converts backtest result to pandas data frames.

    Data will be re-indexed against the predictions' natural time index, dropping any target data that doesn't
    correspond to a prediction.

    Args:
        result_data: result data

    Returns:
        a pandas data frame with the predictions made by the backtest, and their corresponding targets

    Raises:
        NotImplementedError: for result data with more than one target
    """

    # Assume that there is only one target for now.
    if len(result_data["targets"]) != 1:
        raise NotImplementedError

    # Aggregate all predictions into a list of data frames, setting the index to both `time` and `reference_time`.
    prediction_pandas_objects = []
    for fold in result_data["fold_results"]:
        for predict_result in fold["predict_results"]:
            for prediction in predict_result["predictions"]:
                prediction_time_dataset = TimeDataset.parse_obj(prediction)
                # Disable an argument deprecated warning.
                with warnings.catch_warnings():
                    warnings.filterwarnings(
                        "ignore",
                        category=FutureWarning,
                        message="Argument `closed` is deprecated in favor of `inclusive`.",
                    )
                    prediction_pandas_object = prediction_time_dataset.flatten().to_pandas_object()

                multi_index = pd.MultiIndex.from_product(
                    [prediction_pandas_object.index, [pd.Timestamp(predict_result["reference_time"])]],
                    names=["time", "reference_time"],
                )
                prediction_pandas_objects.append(prediction_pandas_object.set_axis(multi_index))

    # Create the prediction data frame.
    prediction_pandas_object = pd.concat(prediction_pandas_objects)

    # Create the target data frame.
    with warnings.catch_warnings():
        warnings.filterwarnings(
            "ignore", category=FutureWarning, message="Argument `closed` is deprecated in favor of `inclusive`."
        )
        target_pandas_object = TimeDataset.parse_obj(result_data["targets"][0]).flatten().to_pandas_object()

    # Re-index the target data frame against the `time` index of the predictions, ensuring that we can make a
    # combined data frame with a single natural time index.
    target_pandas_object = target_pandas_object.reindex(
        prediction_pandas_object.index.get_level_values("time")
    ).set_axis(prediction_pandas_object.index)

    # Cast `target_pandas_object` as a `pd.DataFrame` if `prediction_pandas_object` is a `pd.DataFrame`. Then add
    # empty string columns to the `target_pandas_object` dataframe. We do this
    # to cleanly merge `prediction_pandas_object`'s `pd.MultiIndex` columns.
    if isinstance(prediction_pandas_object, pd.DataFrame) and isinstance(target_pandas_object, pd.Series):
        target_column_names = tuple("" for _ in range(prediction_pandas_object.columns.nlevels))
        target_pandas_object = target_pandas_object.to_frame(name=target_column_names)
        target_pandas_object.columns.names = prediction_pandas_object.columns.names

    # Return the combined data frame.
    pandas_data_frame = pd.concat(
        [target_pandas_object, prediction_pandas_object], axis=1, keys=["targets", "predictions"]
    )

    return pandas_data_frame



from myst.client import Client
from myst.openapi.models.time_series_list import TimeSeriesList


def request_sync(client: Client) -> TimeSeriesList:
    """Lists time series across all projects."""

    return client.request(method="get", path=f"/projects/-/time_series/", response_class=TimeSeriesList)



import enum
from typing import Optional, Sequence, Union
from uuid import UUID

from myst.connectors.model_connector import ModelConnector
from myst.core.time.time_delta import TimeDelta


@enum.unique
class GroupName(str, enum.Enum):
    FEATURES = "features"
    TARGETS = "targets"


@enum.unique
class LabelIndexer(str, enum.Enum):
    LOC = "loc"
    SCALE = "scale"


class ARIMAX(ModelConnector):
    def __init__(
        self,
        run_inputs_time_delta: Optional[TimeDelta] = None,
        auto_regressive_orders: Optional[Union[int, Sequence[int]]] = None,
        integration_order: Optional[int] = None,
        moving_average_orders: Optional[Union[int, Sequence[int]]] = None,
        trend_polynomial_terms: Optional[Sequence[bool]] = None,
        measurement_error: Optional[bool] = None,
        time_varying_regression: Optional[bool] = None,
        mle_regression: Optional[bool] = None,
        enforce_stationarity: Optional[bool] = None,
        enforce_invertibility: Optional[bool] = None,
        hamilton_representation: Optional[bool] = None,
    ):
        super().__init__(
            uuid=UUID("a921bb8c-2699-4dd4-ad8e-a5f32d02555a"),
            parameters=dict(
                run_inputs_time_delta=run_inputs_time_delta,
                auto_regressive_orders=auto_regressive_orders,
                integration_order=integration_order,
                moving_average_orders=moving_average_orders,
                trend_polynomial_terms=trend_polynomial_terms,
                measurement_error=measurement_error,
                time_varying_regression=time_varying_regression,
                mle_regression=mle_regression,
                enforce_stationarity=enforce_stationarity,
                enforce_invertibility=enforce_invertibility,
                hamilton_representation=hamilton_representation,
            ),
        )



from myst.client import Client
from myst.openapi.models.time_series_get import TimeSeriesGet


def request_sync(client: Client, project_uuid: str, uuid: str) -> TimeSeriesGet:
    """Deletes a time series."""

    return client.request(
        method="delete", path=f"/projects/{project_uuid}/time_series/{uuid}", response_class=TimeSeriesGet
    )



import enum
from uuid import UUID

from myst.connectors.model_connector import ModelConnector


@enum.unique
class GroupName(str, enum.Enum):
    FEATURES = "features"
    TARGETS = "targets"
    SAMPLE_WEIGHTS = "sample_weights"


class LinearRegression(ModelConnector):
    def __init__(self, fit_intercept: bool = True, normalize: bool = False) -> None:
        super().__init__(
            uuid=UUID("7b714851-bfbf-46a2-a8e5-8af90b8bae71"),
            parameters=dict(fit_intercept=fit_intercept, normalize=normalize),
        )



from typing import Optional

from myst.models import base_model
from myst.openapi.models.hpo_trial_metrics import HPOTrialMetrics
from myst.openapi.models.hpo_trial_parameters import HPOTrialParameters


class HPOTrial(base_model.BaseModel):
    """Represents the result of a single HPO trial."""

    parameters: HPOTrialParameters
    metrics: HPOTrialMetrics
    create_time: str
    backtest_result_url: Optional[str] = None



from typing import Optional
from uuid import UUID

from myst.core.time.time import Time
from myst.models.base_model import BaseModel


class Resource(BaseModel):
    """Represents an API resource with a persistent identity.

    Attributes:
        uuid: the universally unique identifier for this resource
        create_time: the time when the resource was created
        update_time: the time when the resource was last modified
    """

    uuid: UUID
    create_time: Time
    update_time: Optional[Time] = None


class ShareableResource(Resource):
    """A resource associated with an owner, organization, and creator.

    Attributes:
        organization: the identifier of the organization this resource belongs to
        owner: the identifier of the user who owns this resource
        creator: the identifier of the user who created this resource
    """

    organization: UUID
    owner: UUID
    creator: UUID



import enum
from typing import Optional
from uuid import UUID

from myst.connectors.model_connector import ModelConnector


@enum.unique
class GroupName(str, enum.Enum):
    FEATURES = "features"
    TARGETS = "targets"
    SAMPLE_WEIGHTS = "sample_weights"


@enum.unique
class LogisticRegressionPenalty(str, enum.Enum):
    ELASTIC_NET = "elasticnet"
    L1 = "l1"
    L2 = "l2"
    NONE = "none"


@enum.unique
class LogisticRegressionSolver(str, enum.Enum):
    LBFGS = "lbfgs"
    LIBLINEAR = "liblinear"
    NEWTON_CG = "newton-cg"
    SAG = "sag"
    SAGA = "saga"


class LogisticRegression(ModelConnector):
    def __init__(
        self,
        penalty: Optional[LogisticRegressionPenalty] = None,
        constraint: Optional[float] = None,
        fit_intercept: Optional[bool] = None,
        solver: Optional[LogisticRegressionSolver] = None,
        max_iter: Optional[int] = None,
        l1_ratio: Optional[float] = None,
    ) -> None:
        super().__init__(
            uuid=UUID("789b7a82-2ae4-4a9c-8078-dfa9d14b120b"),
            parameters=dict(
                penalty=penalty,
                constraint=constraint,
                fit_intercept=fit_intercept,
                solver=solver,
                max_iter=max_iter,
                l1_ratio=l1_ratio,
            ),
        )



from myst.client import Client
from myst.openapi.models.hpo_get_list import HPOGetList


def request_sync(client: Client, project_uuid: str) -> HPOGetList:
    """Lists all HPOs in the project."""

    return client.request(method="get", path=f"/projects/{project_uuid}/hpos/-/", response_class=HPOGetList)



from myst.client import Client
from myst.openapi.models.layer_get import LayerGet


def request_sync(client: Client, project_uuid: str, time_series_uuid: str, uuid: str) -> LayerGet:
    """Gets a specific layer for a time series."""

    return client.request(
        method="get",
        path=f"/projects/{project_uuid}/time_series/{time_series_uuid}/layers/{uuid}",
        response_class=LayerGet,
    )



from myst.connectors.source_connectors.cleaned_observations import CleanedObservations
from myst.connectors.source_connectors.enhanced_forecast import EnhancedForecast
from myst.connectors.source_connectors.historical_hourly_conditions import HistoricalHourlyConditions
from myst.connectors.source_connectors.holidays import Holidays
from myst.connectors.source_connectors.solar_position import SolarPosition
from myst.connectors.source_connectors.time_trends import TimeTrends
from myst.connectors.source_connectors.yes_energy import YesEnergy

__all__ = [
    "CleanedObservations",
    "EnhancedForecast",
    "HistoricalHourlyConditions",
    "Holidays",
    "SolarPosition",
    "TimeTrends",
    "YesEnergy",
]



from myst.client import Client
from myst.openapi.models.node_run_job_get import NodeRunJobGet


def request_sync(client: Client, project_uuid: str, time_series_uuid: str, uuid: str) -> NodeRunJobGet:
    """Gets a time series run job by its unique identifier."""

    return client.request(
        method="get",
        path=f"/projects/{project_uuid}/time_series/{time_series_uuid}/run_jobs/{uuid}",
        response_class=NodeRunJobGet,
    )



import enum
from uuid import UUID

from myst.connectors.operation_connector import OperationConnector


@enum.unique
class GroupName(str, enum.Enum):
    OPERANDS = "operands"


class FillNA(OperationConnector):
    def __init__(self, value: float) -> None:
        super().__init__(uuid=UUID("f268a5d7-dc66-4dc8-a053-4d50e9866611"), parameters=dict(value=value))



from uuid import UUID

import myst
from myst.connectors.model_connectors import xgboost
from myst.connectors.operation_connectors import time_transformations
from myst.connectors.source_connector import SourceConnector
from myst.connectors.source_connectors import cleaned_observations, enhanced_forecast, historical_hourly_conditions

myst.authenticate()

project = myst.Project.create(title="SF Electricity Load")

ksfo_cleaned_observations = project.create_source(
    title="Cleaned Weather (KSFO)",
    connector=cleaned_observations.CleanedObservations(
        latitude=37.619,
        longitude=-122.374,
        fields=[
            cleaned_observations.Field.RELATIVE_HUMIDITY_PERCENT,
            cleaned_observations.Field.SURFACE_TEMPERATURE_CELSIUS,
        ],
    ),
)

ksfo_hhc = project.create_source(
    title="Historical Weather (KSFO)",
    connector=historical_hourly_conditions.HistoricalHourlyConditions(
        latitude=37.619,
        longitude=-122.374,
        fields=[historical_hourly_conditions.Field.RELATIVE_HUMIDITY, historical_hourly_conditions.Field.TEMPERATURE],
    ),
)

# If a connector isn't present in the library yet, you can still specify it and its parameters. This is The Weather
# Company's Enhanced Forecast API.
ksfo_forecast = project.create_source(
    title="Forecasted Weather (KSFO)",
    connector=SourceConnector(
        uuid=UUID("1aeaadec-6912-422a-a4fc-5d6765391c17"),
        parameters=dict(latitude=37.619, longitude=-122.374, fields=["relativeHumidity", "temperature"]),
    ),
)

ksfo_temperature_ts = project.create_time_series(title="Temperature (KSFO)", sample_period=myst.TimeDelta("PT1H"))

# Add the three weather sources to the temperature time series.
ksfo_temperature_ts.create_layer(
    node=ksfo_cleaned_observations,
    order=0,
    end_timing=myst.TimeDelta("-PT23H"),
    label_indexer=cleaned_observations.Field.SURFACE_TEMPERATURE_CELSIUS.value,
)
ksfo_temperature_ts.create_layer(
    node=ksfo_hhc,
    order=1,
    start_timing=myst.TimeDelta("-PT23H"),
    end_timing=myst.TimeDelta("PT1H"),
    label_indexer=historical_hourly_conditions.Field.TEMPERATURE.value,
)
ksfo_temperature_ts.create_layer(
    node=ksfo_forecast,
    order=2,
    start_timing=myst.TimeDelta("PT1H"),
    label_indexer=enhanced_forecast.Field.TEMPERATURE.value,
)

assert [layer.upstream_node for layer in ksfo_temperature_ts.list_layers()] == [
    ksfo_cleaned_observations.uuid,
    ksfo_hhc.uuid,
    ksfo_forecast.uuid,
]

# Create a 3-hour rolling mean of temperature.
rolling_mean_op = project.create_operation(
    title="Temperature (KSFO) - 3H Rolling Mean",
    connector=time_transformations.TimeTransformations(
        rolling_window_parameters=time_transformations.RollingWindowParameters(
            window_period=myst.TimeDelta("PT3H"),
            min_periods=1,
            centered=False,
            aggregation_function=time_transformations.AggregationFunction.MEAN,
        )
    ),
)

rolling_mean_op.create_input(time_series=ksfo_temperature_ts, group_name=time_transformations.GroupName.OPERANDS)

ksfo_rolling_mean_temp = project.create_time_series(
    title="Temperature (KSFO) - Rolling 3H Mean", sample_period=myst.TimeDelta("PT1H")
)

ksfo_rolling_mean_temp.create_layer(node=rolling_mean_op, order=0)

historical_sf_electricity_load = project.create_time_series(
    title="Historical SF Electricity Load",
    sample_period=myst.TimeDelta("PT1H"),
    description="Historical SF electricity load (target), to be uploaded into the platform",
)

xgboost_model = project.create_model(title="XGBoost Model", connector=xgboost.XGBoost())

xgboost_model.create_input(time_series=ksfo_rolling_mean_temp, group_name=xgboost.GroupName.FEATURES)

xgboost_model.create_input(time_series=historical_sf_electricity_load, group_name=xgboost.GroupName.TARGETS)

forecast_sf_electricity_load = project.create_time_series(
    title="Forecasted SF Electricity Load", sample_period=myst.TimeDelta("PT1H")
)

forecast_sf_electricity_load.create_layer(node=xgboost_model, order=0)

xgboost_model_fit_policy = xgboost_model.create_fit_policy(
    schedule_timing=myst.TimeDelta("PT24H"),
    start_timing=myst.Time("2021-01-01T00:00:00Z"),
    end_timing=myst.TimeDelta("-PT1H"),
)

forecast_prediction_policy = forecast_sf_electricity_load.create_run_policy(
    schedule_timing=myst.TimeDelta("PT1H"), start_timing=myst.TimeDelta("PT1H"), end_timing=myst.TimeDelta("P7D")
)



from myst.client import Client
from myst.openapi.models.operation_create import OperationCreate
from myst.openapi.models.operation_get import OperationGet


def request_sync(client: Client, project_uuid: str, json_body: OperationCreate) -> OperationGet:
    """Creates an operation."""

    return client.request(
        method="post",
        path=f"/projects/{project_uuid}/operations/",
        response_class=OperationGet,
        request_model=json_body,
    )



from typing import List, Union

from myst.models import base_model
from myst.openapi.models.model_get import ModelGet
from myst.openapi.models.operation_get import OperationGet
from myst.openapi.models.source_get import SourceGet
from myst.openapi.models.time_series_get import TimeSeriesGet


class ProjectNodesList(base_model.BaseModel):
    """Project nodes list schema."""

    data: List[Union[TimeSeriesGet, SourceGet, OperationGet, ModelGet]]
    has_more: bool



from typing import Any, List, Optional

from pydantic import Field
from typing_extensions import Literal

from myst.models import base_model
from myst.openapi.models.deploy_status import DeployStatus


class TimeSeriesGet(base_model.BaseModel):
    """Schema for time series get responses."""

    object_: Literal["Node"] = Field(..., alias="object")
    uuid: str
    create_time: str
    organization: str
    owner: str
    type: Literal["TimeSeries"]
    title: str
    project: str
    creator: str
    deploy_status: DeployStatus
    sample_period: str
    cell_shape: List[Any]
    coordinate_labels: List[Any]
    axis_labels: List[Any]
    update_time: Optional[str] = None
    description: Optional[str] = None



from myst.client import Client
from myst.openapi.models.project_node_policy_list import ProjectNodePolicyList


def request_sync(client: Client, project_uuid: str) -> ProjectNodePolicyList:
    """Lists all policies for a project."""

    return client.request(
        method="get", path=f"/projects/{project_uuid}/nodes/-/policies/", response_class=ProjectNodePolicyList
    )



from typing import List

from myst.models import base_model
from myst.openapi.models.model_fit_result_list_item import ModelFitResultListItem


class ModelFitResultList(base_model.BaseModel):
    """Schema for model fit result list responses."""

    data: List[ModelFitResultListItem]
    has_more: bool



from typing import List

from myst.models import base_model
from myst.openapi.models.time_series_run_policy_get import TimeSeriesRunPolicyGet


class TimeSeriesRunPolicyList(base_model.BaseModel):
    """Schema for time series run policy list responses."""

    data: List[TimeSeriesRunPolicyGet]
    has_more: bool



from typing import Any, Dict, Optional

from myst.client import Client
from myst.openapi.models.time_series_query_result_get import TimeSeriesQueryResultGet


def request_sync(
    client: Client,
    project_uuid: str,
    uuid: str,
    start_time: str,
    end_time: str,
    as_of_time: Optional[str] = None,
    as_of_offset: Optional[str] = None,
) -> TimeSeriesQueryResultGet:
    """Queries time series data."""

    params: Dict[str, Any] = {}
    params["start_time"] = start_time
    params["end_time"] = end_time
    params["as_of_time"] = as_of_time
    params["as_of_offset"] = as_of_offset

    params = {k: v for k, v in params.items() if v is not None}

    return client.request(
        method="get",
        path=f"/projects/{project_uuid}/time_series/{uuid}:query",
        response_class=TimeSeriesQueryResultGet,
        params=params,
    )



from typing import List

from myst.models import base_model
from myst.openapi.models.deployment_get import DeploymentGet


class DeploymentList(base_model.BaseModel):
    """Schema for deployment list responses."""

    data: List[DeploymentGet]
    has_more: bool



from myst.client import Client
from myst.openapi.models.alias_get import AliasGet


def request_sync(client: Client, alias_uuid: str, node_uuid: str) -> AliasGet:
    """Update an alias to point to a different node."""

    return client.request(method="post", path=f"/alias/{alias_uuid}/node/{node_uuid}", response_class=AliasGet)



from myst.client import Client
from myst.openapi.models.model_fit_policy_create import ModelFitPolicyCreate
from myst.openapi.models.model_fit_policy_get import ModelFitPolicyGet


def request_sync(
    client: Client, project_uuid: str, model_uuid: str, json_body: ModelFitPolicyCreate
) -> ModelFitPolicyGet:
    """Creates a model fit policy."""

    return client.request(
        method="post",
        path=f"/projects/{project_uuid}/models/{model_uuid}/fit_policies/",
        response_class=ModelFitPolicyGet,
        request_model=json_body,
    )



from typing import Optional

from pydantic import Field
from typing_extensions import Literal

from myst.models import base_model


class QLogUniform(base_model.BaseModel):
    """"""

    object_: Literal["Sampler"] = Field(..., alias="object")
    type: Literal["QLogUniform"]
    lower: float
    upper: float
    q: float
    base: Optional[int] = 10



import enum
from typing import List
from uuid import UUID

from myst.connectors.source_connector import SourceConnector
from myst.core.time.time_delta import TimeDelta


@enum.unique
class Field(str, enum.Enum):
    CALENDAR_YEAR = "CALENDAR_YEAR"
    DAY_OF_WEEK = "DAY_OF_WEEK"
    DAY_OF_YEAR = "DAY_OF_YEAR"
    HOUR_OF_DAY = "HOUR_OF_DAY"
    IS_WEEKDAY = "IS_WEEKDAY"
    MONTH_OF_YEAR = "MONTH_OF_YEAR"
    WEEK_OF_YEAR = "WEEK_OF_YEAR"
    EPOCH = "EPOCH"


class TimeTrends(SourceConnector):
    def __init__(self, sample_period: TimeDelta, time_zone: str, fields: List[Field]) -> None:
        super().__init__(
            uuid=UUID("7621e074-3d62-4735-911d-91a2a971240d"),
            parameters=dict(sample_period=sample_period, time_zone=time_zone, fields=fields),
        )



from myst.client import Client
from myst.openapi.models.hpo_job_get_list import HPOJobGetList


def request_sync(client: Client, project_uuid: str) -> HPOJobGetList:
    """Lists all HPO jobs for the project."""

    return client.request(method="get", path=f"/projects/{project_uuid}/hpos/-/jobs/", response_class=HPOJobGetList)



from myst.client import Client
from myst.openapi.models.hpo_job_get import HPOJobGet


def request_sync(client: Client, project_uuid: str, uuid: str) -> HPOJobGet:
    """Runs an HPO job."""

    return client.request(method="post", path=f"/projects/{project_uuid}/hpos/{uuid}:run", response_class=HPOJobGet)



from myst.client import Client
from myst.openapi.models.layer_list import LayerList


def request_sync(client: Client, project_uuid: str, time_series_uuid: str) -> LayerList:
    """Lists layers for a time series."""

    return client.request(
        method="get", path=f"/projects/{project_uuid}/time_series/{time_series_uuid}/layers/", response_class=LayerList
    )



from myst.connectors.model_connectors.arimax import ARIMAX
from myst.connectors.model_connectors.elastic_net import ElasticNet
from myst.connectors.model_connectors.extra_trees_regression import ExtraTreesRegression
from myst.connectors.model_connectors.lightgbm import LightGBM
from myst.connectors.model_connectors.linear_regression import LinearRegression
from myst.connectors.model_connectors.logistic_regression import LogisticRegression
from myst.connectors.model_connectors.mlp_regression import MLPRegression
from myst.connectors.model_connectors.ngboost import NGBoost
from myst.connectors.model_connectors.random_forest_regression import RandomForestRegression
from myst.connectors.model_connectors.xgboost import XGBoost

__all__ = [
    "ElasticNet",
    "ExtraTreesRegression",
    "LightGBM",
    "LinearRegression",
    "LogisticRegression",
    "MLPRegression",
    "NGBoost",
    "RandomForestRegression",
    "ARIMAX",
    "XGBoost",
]



from typing import Optional

from pydantic import Field
from typing_extensions import Literal

from myst.models import base_model
from myst.openapi.models.hpo_metric import HPOMetric


class HyperoptGet(base_model.BaseModel):
    """Hyperopt search algorithm schema for get responses.

    See https://hyperopt.github.io/hyperopt/ for `hyperopt` reference."""

    metric: HPOMetric
    num_trials: int
    max_concurrent_trials: int
    algorithm: Literal["TPE"]
    n_startup_jobs: int
    object_: Optional[Literal["SearchAlgorithm"]] = Field(..., alias="object")
    type: Optional[Literal["Hyperopt"]] = "Hyperopt"



import enum
from uuid import UUID

from myst.connectors.source_connector import SourceConnector
from myst.core.time.time_delta import TimeDelta


@enum.unique
class CountryField(str, enum.Enum):
    ANGOLA = "Angola"
    ARGENTINA = "Argentina"
    ARUBA = "Aruba"
    AUSTRALIA = "Australia"
    AUSTRIA = "Austria"
    BANGLADESH = "Bangladesh"
    BELARUS = "Belarus"
    BELGIUM = "Belgium"
    BOTSWANA = "Botswana"
    BRAZIL = "Brazil"
    BULGARIA = "Bulgaria"
    BURUNDI = "Burundi"
    CANADA = "Canada"
    CHILE = "Chile"
    COLOMBIA = "Colombia"
    CROATIA = "Croatia"
    CURACAO = "Curacao"
    CZECHIA = "Czechia"
    DENMARK = "Denmark"
    DJIBOUTI = "Djibouti"
    DOMINICAN_REPUBLIC = "DominicanRepublic"
    EGYPT = "Egypt"
    ENGLAND = "England"
    ESTONIA = "Estonia"
    EUROPEAN_CENTRAL_BANK = "EuropeanCentralBank"
    FINLAND = "Finland"
    FRANCE = "France"
    GEORGIA = "Georgia"
    GERMANY = "Germany"
    GREECE = "Greece"
    HONDURAS = "Honduras"
    HONGKONG = "HongKong"
    HUNGARY = "Hungary"
    ICELAND = "Iceland"
    INDIA = "India"
    IRELAND = "Ireland"
    ISLE_OF_MAN = "IsleOfMan"
    ISRAEL = "Israel"
    ITALY = "Italy"
    JAMAICA = "Jamaica"
    JAPAN = "Japan"
    KENYA = "Kenya"
    KOREA = "Korea"
    LATVIA = "Latvia"
    LITHUANIA = "Lithuania"
    LUXEMBOURG = "Luxembourg"
    MALAYSIA = "Malaysia"
    MALAWI = "Malawi"
    MEXICO = "Mexico"
    MOROCCO = "Morocco"
    MOZAMBIQUE = "Mozambique"
    NETHERLANDS = "Netherlands"
    NEW_ZEALAND = "NewZealand"
    NICARAGUA = "Nicaragua"
    NIGERIA = "Nigeria"
    NORTHERN_IRELAND = "NorthernIreland"
    NORWAY = "Norway"
    PARAGUAY = "Paraguay"
    PERU = "Peru"
    POLAND = "Poland"
    PORTUGAL = "Portugal"
    PORTUGAL_EXTENDED = "PortugalExt"
    ROMANIA = "Romania"
    RUSSIA = "Russia"
    SAUDI_ARABIA = "SaudiArabia"
    SCOTLAND = "Scotland"
    SERBIA = "Serbia"
    SINGAPORE = "Singapore"
    SLOVAKIA = "Slovakia"
    SLOVENIA = "Slovenia"
    SOUTH_AFRICA = "SouthAfrica"
    SPAIN = "Spain"
    SWEDEN = "Sweden"
    SWITZERLAND = "Switzerland"
    TURKEY = "Turkey"
    UKRAINE = "Ukraine"
    UNITED_ARAB_EMIRATES = "UnitedArabEmirates"
    UNITED_KINGDOM = "UnitedKingdom"
    UNITED_STATES = "UnitedStates"
    VENEZUELA = "Venezuela"
    VIETNAM = "Vietnam"
    WALES = "Wales"


class Holidays(SourceConnector):
    def __init__(self, sample_period: TimeDelta, time_zone: str, country: CountryField) -> None:
        super().__init__(
            uuid=UUID("7c712128-ff20-4a71-b69d-92c5b3df665e"),
            parameters=dict(sample_period=sample_period, time_zone=time_zone, country=country),
        )



"""This module contains the `UserCredentialsCache` class."""
import json
import os
from pathlib import Path
from typing import Optional

from google.oauth2.credentials import Credentials as GoogleCredentials


class UserCredentialsCache:
    """User credentials cache class for saving and loading user credentials to a local cache.

    This writes user credentials to a subdirectory within the `~/.config` directory (or `APPDATA` on Windows).
    """

    _CACHE_DIRNAME = "myst"
    _USER_CREDENTIALS_FILENAME = "myst_user_credentials.json"
    _USER_HOME_DIR = Path("~").expanduser()
    _CONFIG_HIDDEN = ".config"

    @property
    def cache_file_dir(self) -> Path:
        """Directory where we'll save the cached credentials."""
        if os.name == "nt":  # needed for Windows.
            return Path(os.environ["APPDATA"])
        else:
            return self._USER_HOME_DIR / self._CONFIG_HIDDEN

    @property
    def cache_file_path(self) -> Path:
        """Where to cache the user credentials in the local filesystem."""
        return self.cache_file_dir / self._CACHE_DIRNAME / self._USER_CREDENTIALS_FILENAME

    def load(self) -> Optional[GoogleCredentials]:
        """Loads credentials from cache.

        Returns:
            user_credentials or None if we could not not load them for some reason (eg. they didn't exist, they were
                malformed, etc.)
        """
        if not self.cache_file_path.exists():
            return None

        with self.cache_file_path.open("r") as user_credentials_file:
            user_credentials_json = json.load(user_credentials_file)

        # Note that we don't save the OAuth 2.0 access token, so we're explicitly setting it to None here.
        user_credentials = GoogleCredentials(token=None, **user_credentials_json)
        return user_credentials

    def save(self, user_credentials: GoogleCredentials) -> None:
        """Saves user credentials to cache."""
        cache_dir_path = self.cache_file_path.parent
        if not cache_dir_path.exists():
            cache_dir_path.mkdir(parents=True)

        # Note that we don't save the OAuth 2.0 access token, so we're explicitly not passing it in here.
        user_credentials_json = {
            "refresh_token": user_credentials.refresh_token,
            "id_token": user_credentials.id_token,
            "token_uri": user_credentials.token_uri,
            "client_id": user_credentials.client_id,
            "client_secret": user_credentials.client_secret,
            "scopes": user_credentials.scopes,
        }

        with self.cache_file_path.open("w") as cache_file:
            json.dump(user_credentials_json, cache_file)

    def clear(self) -> None:
        """Clears the user credentials cache."""
        self.cache_file_path.unlink()



from typing import List, Union

from pydantic import Field
from typing_extensions import Literal

from myst.models import base_model


class Choice(base_model.BaseModel):
    """"""

    object_: Literal["Sampler"] = Field(..., alias="object")
    type: Literal["Choice"]
    choices: List[Union[float, int, str]]



from myst.client import Client
from myst.openapi.models.organization_get import OrganizationGet


def request_sync(client: Client, uuid: str) -> OrganizationGet:
    """Gets an organization by its unique identifier."""

    return client.request(method="get", path=f"/organizations/{uuid}", response_class=OrganizationGet)



from typing import List

from myst.models import base_model
from myst.openapi.models.hpo_result_get import HPOResultGet


class HPOResultList(base_model.BaseModel):
    """Schema for model fit result list responses."""

    data: List[HPOResultGet]
    has_more: bool



from myst.client import Client
from myst.openapi.models.connector_get import ConnectorGet


def request_sync(client: Client, uuid: str) -> ConnectorGet:
    """Gets a connector by its unique identifier."""

    return client.request(method="get", path=f"/connectors/{uuid}", response_class=ConnectorGet)



from typing import List

from myst.models import base_model
from myst.openapi.models.hpo_job_get import HPOJobGet


class HPOJobGetList(base_model.BaseModel):
    """Schema for HPOJobGet list responses."""

    data: List[HPOJobGet]
    has_more: bool



from typing import Any, Dict, Union

from myst.models import base_model
from myst.openapi.models.choice import Choice
from myst.openapi.models.constant import Constant
from myst.openapi.models.log_uniform import LogUniform
from myst.openapi.models.q_log_uniform import QLogUniform
from myst.openapi.models.q_uniform import QUniform
from myst.openapi.models.uniform import Uniform


class SearchSpace(base_model.BaseModel):
    """Represents a search space, or a mapping from parameters to sampling functions."""

    __root__: Dict[str, Union[Choice, Constant, LogUniform, QLogUniform, QUniform, Uniform]]

    def __getitem__(self, item: str) -> Any:
        return self.__root__[item]



import enum
from typing import Optional
from uuid import UUID

from myst.connectors.model_connector import ModelConnector


@enum.unique
class GroupName(str, enum.Enum):
    FEATURES = "features"
    TARGETS = "targets"
    SAMPLE_WEIGHTS = "sample_weights"


@enum.unique
class LabelIndexer(str, enum.Enum):
    LOC = "loc"
    SCALE = "scale"


@enum.unique
class NGBoostDistribution(str, enum.Enum):
    CAUCHY = "Cauchy"
    LAPLACE = "Laplace"
    LOG_NORMAL = "LogNormal"
    NORMAL = "Normal"


class NGBoost(ModelConnector):
    def __init__(
        self,
        distribution: Optional[NGBoostDistribution] = None,
        n_estimators: Optional[int] = None,
        learning_rate: Optional[float] = None,
        minibatch_frac: Optional[float] = None,
        col_sample: Optional[float] = None,
        max_depth: Optional[float] = None,
    ) -> None:
        super().__init__(
            uuid=UUID("7138f288-b588-456f-8164-026d9dea810b"),
            parameters=dict(
                distribution=distribution,
                n_estimators=n_estimators,
                learning_rate=learning_rate,
                minibatch_frac=minibatch_frac,
                col_sample=col_sample,
                max_depth=max_depth,
            ),
        )



from typing import Any, Dict, Optional, Tuple, Union

import numpy as np
import pandas as pd
from pydantic import Field, root_validator, validator

from myst.core.data.validators import (
    check_arrays_same_length_as_time_range,
    ensure_not_none,
    maybe_to_boolean_int_array,
    to_numpy_float_array,
)
from myst.core.time.time import Time, TimeDelta
from myst.models.base_model import BaseModel
from myst.models.types import AxisLabels, CoordinateLabels, ItemOrSlice, Shape

_TIME_INDEX_NAME = "time"


def _process_pandas_object(pandas_object: Union[pd.Series, pd.DataFrame]) -> Tuple[TimeDelta, Time, Time]:
    """Processes the time arguments when the passed data is a `pandas.Series` or `pandas.DataFrame`.

    Args:
        pandas_object: pandas series or data frame

    Returns:
        tuple of (sample period, start time, end time)

    Raises:
        ValueError: if index does not have a frequency or if the pandas object is empty
    """
    # Validate pandas object fulfills `TimeArray` requirements.
    if pandas_object.index.freq is None:
        raise ValueError("The frequency of the time index must be specified.")

    if pandas_object.empty:
        raise ValueError("Creating a time array from an empty Pandas object is not allowed.")

    sample_period = TimeDelta(pandas_object.index.freq)
    start_time = Time(pandas_object.index.min())
    end_time = Time(pandas_object.index.max() + sample_period.to_pandas_date_offset())

    return sample_period, start_time, end_time


def _compute_cell_shape(coordinate_labels: CoordinateLabels) -> Shape:
    """Computes the cell shape from the given coordinate labels.

    Args:
        coordinate_labels: coordinate labels to compute cell shape from

    Returns:
        computed cell shape
    """
    # The cell shape is computed based on the number of coordinates in each dimension.
    return tuple(map(len, coordinate_labels))


def _compute_coordinate_and_axis_labels_for_pandas_index(pandas_index: pd.Index) -> Tuple[CoordinateLabels, AxisLabels]:
    """Computes the coordinate and axis labels from a `pandas.Index`.

    This will typically be used to compute the coordinate and axis labels for a `pandas.DataFrame` given its `columns`
    index.

    Args:
        pandas_index: pandas index to compute the coordinate and axis labels from

    Returns:
        coordinate_labels: coordinate labels
        axis_labels: axis labels
    """
    coordinate_labels: CoordinateLabels = tuple(
        map(tuple, [pandas_index.unique(level=level).tolist() for level in range(pandas_index.nlevels)])
    )
    axis_labels = tuple(pandas_index.names)

    return coordinate_labels, axis_labels


def _get_axis_labels_to_collapse(indexer: ItemOrSlice, axis_labels: AxisLabels) -> AxisLabels:
    """Returns the axis labels that are indexed with scalar indexers, and therefore should be collapsed.

    Args:
        indexer: item or slice; if a tuple of items, one must be present for each axis in `axis_labels`
        axis_labels: axis labels identifying axis over which each item in `indexer` indexes into

    Returns:
        axis labels that were indexed with a scalar value, and therefore should be collapsed
    """
    return tuple(
        axis_label for sub_item, axis_label in zip(indexer, axis_labels) if isinstance(sub_item, (int, str))  # type: ignore
    )


def ndarray_to_list(a: np.ndarray) -> list:
    """Converts a numpy array to a json-serializable list."""
    return np.where(np.isnan(a), None, a).tolist() if a is not None else None


class TimeArray(BaseModel):
    class Config:
        # Allow arbitrary types in order to support `np.ndarray` directly.
        arbitrary_types_allowed = True
        json_encoders = {np.ndarray: ndarray_to_list}

    sample_period: TimeDelta
    start_time: Time
    end_time: Time
    as_of_time: Time
    values: np.ndarray
    mask: Optional[np.ndarray] = None
    coordinate_labels: CoordinateLabels = Field(None)
    axis_labels: AxisLabels = Field(None)

    _ensure_values_not_none = validator("values", pre=True, allow_reuse=True)(ensure_not_none)
    _coerce_values_to_float_array = validator("values", pre=True, allow_reuse=True)(to_numpy_float_array)

    _coerce_mask_to_boolean_int_array = validator("mask", pre=True, allow_reuse=True)(maybe_to_boolean_int_array)

    @validator("coordinate_labels", pre=True, always=True)
    def fill_in_coordinate_labels_if_none(cls, coordinate_labels: Any, values: Dict[str, Any]) -> Any:
        """Returns coordinate labels unchanged if not None, otherwise default integer-valued coordinate labels."""
        ndarray = values.get("values", None)

        if ndarray is not None and coordinate_labels is None:
            if len(ndarray.shape) > 1:
                return tuple([tuple(range(dim)) for dim in tuple(ndarray.shape[1:])])
            else:
                return ()

        return coordinate_labels

    @validator("coordinate_labels")
    def check_coordinate_labels(cls, coordinate_labels: CoordinateLabels, values: Dict[str, Any]) -> CoordinateLabels:
        """Checks that the coordinate labels are unique in each axis and compatible with cell shape of values."""
        ndarray = values.get("values", None)

        if coordinate_labels is not None and ndarray is not None:
            cell_shape = ndarray.shape[1:]
            coordinate_cell_shape = _compute_cell_shape(coordinate_labels=coordinate_labels)

            if cell_shape != coordinate_cell_shape:
                raise ValueError(f"Passed `coordinate_labels` are not compatible with cell shape of `{cell_shape}`.")

            if len(coordinate_cell_shape) > 0:
                if list(map(len, map(set, coordinate_labels))) != list(map(len, coordinate_labels)):  # type: ignore
                    raise ValueError("Passed `coordinate_labels` must contain only unique values across each axis.")

        return coordinate_labels

    @validator("axis_labels", pre=True, always=True)
    def fill_in_axis_labels_if_none(cls, axis_labels: Any, values: Dict[str, Any]) -> Any:
        """Returns axis labels unchanged if not None, otherwise default integer-valued coordinate labels."""
        ndarray = values.get("values", None)

        if ndarray is not None and axis_labels is None:
            if len(ndarray.shape) > 1:
                return tuple(range(len(ndarray.shape[1:])))
            else:
                return ()

        return axis_labels

    @validator("axis_labels")
    def check_axis_labels(cls, axis_labels: AxisLabels, values: Dict[str, Any]) -> AxisLabels:
        """Checks that the axis labels are unique and compatible with cell shape of values."""
        ndarray = values.get("values", None)

        if axis_labels is not None and ndarray is not None:
            cell_shape = ndarray.shape[1:]

            if len(cell_shape) != len(axis_labels):
                raise ValueError(f"Passed `axis_labels` are not compatible with cell shape of `{cell_shape}`.")

            if len(set(axis_labels)) != len(axis_labels):
                raise ValueError("Passed `axis_labels` must contain only unique values.")

        return axis_labels

    @root_validator
    def check_time_range_parameters(cls, values: Dict[str, Any]) -> Dict[str, Any]:
        start_time: Optional[Time] = values.get("start_time", None)
        end_time: Optional[Time] = values.get("end_time", None)
        as_of_time: Optional[Time] = values.get("as_of_time", None)
        sample_period: Optional[TimeDelta] = values.get("sample_period", None)

        if start_time and end_time and as_of_time and sample_period:
            frequency = sample_period.to_pandas_frequency().freqstr
            for time in (start_time, end_time, as_of_time):
                timestamp = time.to_pandas_timestamp()
                if timestamp != timestamp.floor(frequency):
                    raise ValueError(f"Specified time `{time}` is not aligned with frequency `{frequency}`.")

        return values

    @root_validator
    def check_length_of_data_against_index(cls, values: Dict[str, Any]) -> Dict[str, Any]:
        start_time: Optional[Time] = values.get("start_time", None)
        end_time: Optional[Time] = values.get("end_time", None)
        sample_period: Optional[TimeDelta] = values.get("sample_period", None)
        data: Optional[np.ndarray] = values.get("values", None)
        mask: Optional[np.ndarray] = values.get("mask", None)

        if start_time and end_time and sample_period and data is not None:
            check_arrays_same_length_as_time_range(
                start_time=start_time, end_time=end_time, sample_period=sample_period, values=data, mask=mask
            )

        return values

    @property
    def shape(self) -> Shape:
        """The shape of this time array, where the time axis is included as the first dimension."""
        return self.values.shape

    @property
    def cell_shape(self) -> Shape:
        """The shape of the data in this time array at a single timestamp."""
        return self.values.shape[1:]

    @property
    def empty(self) -> bool:
        """Returns True if this time array is empty, False otherwise."""
        return self.values.size == 0

    def __repr__(self) -> str:
        """Returns a representation of the underlying `pandas.Series` or `pandas.DataFrame`."""
        return repr(self.to_pandas_object())

    def __getitem__(self, item: ItemOrSlice) -> "TimeArray":
        """Gets an item or slice from the time array.

        Note that this method only support slicing on the `coordinate_labels`, so not on the time dimension. Time arrays
        with scalar data can never be sliced, since they have no coordinate labels.

        This method will not automatically apply the `mask` to the subset time array, but it will be propagated through
        appropriately in the `TimeArray.mask` attribute.

        The slicing behavior of this method diverges from the behavior of the `loc` method in `pandas` in certain cases.
        The `loc` method only collapses the axes if only scalar indexers are given (e.g. `("a", "z")`), whereas this
        method always collapses axes that correspond to scalar indexers.

        This difference is best illustrated by a few examples. Assume we have a `TimeArray` with the coordinate labels
        `(("a", "b"), ("x", "y", "z"))` and axis labels `("first", "second")`:

            * Indexing `["a"]` will result in the coordinate labels `("x", "y", "z")`. Note that the `"first"` axis is
              collapsed. The `loc` method in `pandas` behaves the same in this case.
            * Indexing `["a", "z"]` will result in the coordinate labels `()`. Note that both axes are collapsed. The
              `loc` method in `pandas` also behaves the same in this case.
            * Indexing using `["a", ["z"]]` – or more generally, including at least one non-scalar indexer – will result
              in the coordinate labels `(("z",),)`. Note that only the `"first"` axis is collapsed. The `loc` method in
              `pandas` *does not* behave the same in this case. Rather, it chooses to not collapse any axes and would
              return coordinate labels `("a", ("z",),)`. We found this behavior inconsistent and chose to improve upon
              it in this method.

        Args:
            item: item or slice

        Returns:
            resultant subset time array

        Raises:
            KeyError: if an invalid item for the time array's shape is given
        """
        if self.cell_shape == ():
            raise KeyError("Cannot slice a scalar time array.")

        # Convert `values` and `mask` `numpy.ndarray` objects into `pandas` objects to take advantage of advanced
        # multi-indexing support.
        subset_values_pandas_object = self.to_pandas_object(apply_mask=False).copy()

        # Note: For some reason, `pandas` will not let you use scalar indexers on a multi-index column with only one
        #   level, so convert the columns into a normal index before slicing into it.
        if len(self.cell_shape) == 1:
            subset_values_pandas_object.columns = subset_values_pandas_object.columns.get_level_values(0)

        # Leverage the already computed multi-index but fill it in with the `mask` data.
        subset_mask_pandas_object = subset_values_pandas_object.copy()
        subset_mask_pandas_object.values[:] = np.zeros(subset_mask_pandas_object.shape)
        if self.mask is not None:
            subset_mask_pandas_object.values[:] = self.mask.reshape(subset_mask_pandas_object.shape)

        # Perform the slice.
        subset_values_pandas_object = subset_values_pandas_object.loc[:, item]
        subset_mask_pandas_object = subset_mask_pandas_object.loc[:, item]

        # If the resultant `subset_values_pandas_object` is empty, that means no matches were found for the slice,
        # making it invalid.
        if subset_values_pandas_object.empty:
            raise KeyError(f"Invalid indexer: `{item}`.")

        # If multiple dimensions are specified and at least one non-scalar item was referenced, remove levels that were
        # referenced via scalar values.
        if isinstance(item, tuple):
            is_scalar_level = [isinstance(sub_item, (int, str)) for sub_item in item]

            if not all(is_scalar_level):
                levels_to_collapse = _get_axis_labels_to_collapse(indexer=item, axis_labels=self.axis_labels)

                subset_values_pandas_object = subset_values_pandas_object.droplevel(axis=1, level=levels_to_collapse)
                subset_mask_pandas_object = subset_mask_pandas_object.droplevel(axis=1, level=levels_to_collapse)

        # Compute the coordinate and axis labels.
        if isinstance(subset_values_pandas_object, pd.DataFrame):
            coordinate_labels, axis_labels = _compute_coordinate_and_axis_labels_for_pandas_index(
                subset_values_pandas_object.columns
            )
        else:
            coordinate_labels, axis_labels = (), ()

        # Compute the new cell shape to use to reshape the two-dimensional `pandas` object `values` to adhere to the
        # `coordinate_labels`.
        cell_shape = _compute_cell_shape(coordinate_labels)

        return TimeArray(
            sample_period=self.sample_period,
            start_time=self.start_time,
            end_time=self.end_time,
            as_of_time=self.as_of_time,
            values=subset_values_pandas_object.values.reshape(-1, *cell_shape),
            mask=subset_mask_pandas_object.values.astype(np.int8).reshape(-1, *cell_shape),
            coordinate_labels=coordinate_labels,
            axis_labels=axis_labels,
        )

    @classmethod
    def from_numpy_masked_array(
        cls,
        numpy_masked_array: np.ma.MaskedArray,
        sample_period: TimeDelta,
        start_time: Time,
        end_time: Time,
        as_of_time: Time,
        coordinate_labels: Optional[CoordinateLabels] = None,
        axis_labels: Optional[AxisLabels] = None,
    ) -> "TimeArray":
        """Constructs a time array from a `numpy.ma.MaskedArray` instance.

        Args:
            numpy_masked_array: numpy masked array
            sample_period: sample period
            start_time: start time
            end_time: end time
            as_of_time: as of time
            coordinate_labels: labels for individual coordinates
            axis_labels: names for individual axes

        Returns:
            time array
        """
        return TimeArray(
            sample_period=sample_period,
            start_time=start_time,
            end_time=end_time,
            as_of_time=as_of_time,
            values=numpy_masked_array.data,
            mask=np.array(numpy_masked_array.mask, dtype=int) if numpy_masked_array.mask is not False else None,
            coordinate_labels=coordinate_labels,
            axis_labels=axis_labels,
        )

    @classmethod
    def from_pandas_data_frame(
        cls, pandas_data_frame: pd.DataFrame, as_of_time: Time, mask: Optional[np.ndarray] = None
    ) -> "TimeArray":
        """Constructs a time array from a `pandas.DataFrame` instance.

        In order to use this constructor, the `pandas.DataFrame` instance must be indexed by a `pandas.DatetimeIndex`
        with a specified frequency which is used to define the `sample_period` of the `TimeArray`.

        Additionally, the `columns` must be instantiated with a named `pandas.Index`, which is used to define the
        `axis_labels` and `coordinate_labels` of the `TimeArray`.

        Args:
            pandas_data_frame: pandas data frame
            as_of_time: as of time
            mask: binary array with mask

        Returns:
            time array

        Raises:
            ValueError: if data frame does not represent a valid time array
        """
        sample_period, start_time, end_time = _process_pandas_object(pandas_object=pandas_data_frame)

        # Extract the coordinate and axis labels from the columns of the `pandas.DataFrame`.
        coordinate_labels, axis_labels = _compute_coordinate_and_axis_labels_for_pandas_index(
            pandas_index=pandas_data_frame.columns
        )

        # Ensure that the coordinate labels are integers or strings.
        for labels in coordinate_labels:
            for label in labels:
                if not isinstance(label, (int, str)):
                    raise ValueError("All coordinate labels must be integers or strings.")

        # Ensure that the axis labels are integers or strings.
        for axis_label in axis_labels:
            if not isinstance(axis_label, (int, str)):
                raise ValueError(
                    "Data frame columns must be instantiated with a named index to use as `axis_labels`. Names must "
                    "be integers or strings."
                )

        # Compute the cell shape given the coordinate labels.
        cell_shape = _compute_cell_shape(coordinate_labels=coordinate_labels)

        # Since `pandas.DataFrame.values` only holds two-dimensional data, reshape the data to adhere to the
        # `coordinate_labels`.
        values = pandas_data_frame.values.reshape(-1, *cell_shape)

        return TimeArray(
            sample_period=sample_period,
            start_time=start_time,
            end_time=end_time,
            as_of_time=as_of_time,
            values=values,
            mask=mask,
            coordinate_labels=coordinate_labels,
            axis_labels=axis_labels,
        )

    @classmethod
    def from_pandas_series(
        cls, pandas_series: pd.Series, as_of_time: Time, mask: Optional[np.ndarray] = None
    ) -> "TimeArray":
        """Constructs a time array from a `pandas.Series` instance.

        In order to use this constructor, the `pandas.Series` instance must be indexed by a `pandas.DatetimeIndex` with
        a specified frequency which is used to define the `sample_period` of the `TimeArray`.

        Additionally, the series must be named with an integer or string, which is used to define the
        `coordinate_labels` of the `TimeArray`.

        Args:
            pandas_series: pandas series
            as_of_time: as of time
            mask: binary array with mask

        Returns:
            time array
        """
        sample_period, start_time, end_time = _process_pandas_object(pandas_object=pandas_series)

        return TimeArray(
            sample_period=sample_period,
            start_time=start_time,
            end_time=end_time,
            as_of_time=as_of_time,
            values=pandas_series.values,
            mask=mask,
        )

    @classmethod
    def from_pandas_object(
        cls, pandas_object: Union[pd.Series, pd.DataFrame], as_of_time: Time, mask: Optional[np.ndarray] = None
    ) -> "TimeArray":
        """Constructs a time array from a `pandas.Series` or `pandas.DataFrame` instance.

        In order to use this constructor, the pandas object must be indexed by a `pandas.DatetimeIndex` with
        a specified frequency which is used to define the `sample_period` of the `TimeArray`.

        Additionally, if the object is a series, it must be named with an integer or string, which is used to define the
        `coordinate_labels` of the `TimeArray`. If the object is a data frame, the `columns` must be instantiated with
        a named `pandas.Index`, which is used to define the `axis_labels` and `coordinate_labels` of the `TimeArray`.

        Args:
            pandas_object: a pandas series or data frame
            as_of_time: as of time
            mask: binary array with mask

        Returns:
            time array

        Raises:
            ValueError: if object is neither a `pandas.Series` or `pandas.DataFrame`
        """
        if isinstance(pandas_object, pd.Series):
            time_array = cls.from_pandas_series(pandas_series=pandas_object, as_of_time=as_of_time, mask=mask)
        elif isinstance(pandas_object, pd.DataFrame):
            time_array = cls.from_pandas_data_frame(pandas_data_frame=pandas_object, as_of_time=as_of_time, mask=mask)
        else:
            raise ValueError("Object must be a `pandas.Series` or `pandas.DataFrame`.")

        return time_array

    def to_numpy_masked_array(self) -> np.ma.MaskedArray:
        """Returns the `numpy.ma.MaskedArray` equivalent of this time array."""
        return np.ma.masked_array(self.values, mask=self.mask)

    def to_pandas_data_frame(self, apply_mask: Optional[bool] = True) -> pd.DataFrame:
        """Returns the `pandas.DataFrame` equivalent of the values in this time array.

        Note that the `mask` of this time array will need to be extracted separately, as this method does not preserve
        the mask data.

        Args:
            apply_mask: whether or not to apply the internal `mask` before creating the data frame

        Returns:
            data frame

        Raises:
            ValueError: if the cells are dimensionless
        """
        if self.cell_shape == ():
            raise ValueError("Cells need to have one or more dimensions.")

        # Construct the values, optionally applying the internal `mask`.
        values = self.values.copy()
        if apply_mask and self.mask is not None:
            values[self.mask.astype(bool)] = np.nan

        index = pd.date_range(
            start=self.start_time.to_pandas_timestamp(),
            end=self.end_time.to_pandas_timestamp(),
            freq=self.sample_period.to_pandas_frequency(),
            name=_TIME_INDEX_NAME,
            closed="left",
        )

        return pd.DataFrame(
            index=index,
            columns=pd.MultiIndex.from_product(self.coordinate_labels, names=self.axis_labels),
            data=values.reshape(len(values), -1),
        )

    def to_pandas_series(self, apply_mask: Optional[bool] = True) -> pd.Series:
        """The `pandas.Series` equivalent of the `values` in this time array.

        Note that the `mask` of this time array will need to be extracted separately, as this method does not preserve
        the mask data.

        Args:
            apply_mask: whether or not to apply the internal `mask` before creating the series

        Returns:
            pandas series

        Raises:
            ValueError: if the cells are not dimensionless
        """
        if self.cell_shape != ():
            raise ValueError("Cells need to be dimensionless.")

        # Construct the values, optionally applying the internal `mask`.
        values = self.values.copy()
        if apply_mask and self.mask is not None:
            values[self.mask.astype(bool)] = np.nan

        index = pd.date_range(
            start=self.start_time.to_pandas_timestamp(),
            end=self.end_time.to_pandas_timestamp(),
            freq=self.sample_period.to_pandas_frequency(),
            name=_TIME_INDEX_NAME,
            closed="left",
        )

        return pd.Series(index=index, data=values)

    def to_pandas_object(self, apply_mask: Optional[bool] = True) -> Union[pd.Series, pd.DataFrame]:
        """The `pandas.Series` or `pandas.DataFrame` equivalent of the `values` in this time array.

        Note that the `mask` of this time array will need to be extracted separately.

        Args:
            apply_mask: whether or not to apply the internal `mask` before creating the pandas object

        Returns:
            pandas object (series or dataframe)
        """
        if len(self.cell_shape) == 0:
            return self.to_pandas_series(apply_mask=apply_mask)
        else:
            return self.to_pandas_data_frame(apply_mask=apply_mask)



from myst.client import Client
from myst.openapi.models.time_series_run_policy_create import TimeSeriesRunPolicyCreate
from myst.openapi.models.time_series_run_policy_get import TimeSeriesRunPolicyGet


def request_sync(
    client: Client, project_uuid: str, time_series_uuid: str, json_body: TimeSeriesRunPolicyCreate
) -> TimeSeriesRunPolicyGet:
    """Creates a time series run policy."""

    return client.request(
        method="post",
        path=f"/projects/{project_uuid}/time_series/{time_series_uuid}/run_policies/",
        response_class=TimeSeriesRunPolicyGet,
        request_model=json_body,
    )



import pydantic

from myst.core.time.time import Time
from myst.core.time.time_delta import TimeDelta


class BaseModel(pydantic.BaseModel):
    """Base pydantic model to use throughout client library models.

    This model imbues all subclasses with the `json_encoders` specified in its `Config`. This means that types that
    normally wouldn't be serializable will be run through those encoders first.
    """

    class Config:
        json_encoders = {Time: Time.to_iso_string, TimeDelta: TimeDelta.to_iso_string}



from typing import Optional, Union

from pydantic import Field
from typing_extensions import Literal

from myst.models import base_model
from myst.openapi.models.absolute_timing_create import AbsoluteTimingCreate
from myst.openapi.models.relative_timing_create import RelativeTimingCreate


class NodeJobCreate(base_model.BaseModel):
    """Abstract base job schema for create post."""

    object_: Literal["NodeJob"] = Field(..., alias="object")
    start_timing: Union[AbsoluteTimingCreate, RelativeTimingCreate]
    end_timing: Union[AbsoluteTimingCreate, RelativeTimingCreate]
    as_of_time: Optional[AbsoluteTimingCreate] = None



import os
import pathlib
import re
import shutil
from typing import Dict

import typer
from openapi_python_client import Config, MetaType, create_new_client

TOOLS_ROOT = pathlib.Path(__file__).parent.absolute()
PROJECT_ROOT = TOOLS_ROOT.parent.absolute()
HANDWRITTEN_MODELS_DIR = PROJECT_ROOT / "myst" / "models"
_KNOWN_HANDWRITTEN_MODELS = ("time_dataset",)
OPENAPI_URL = "https://api.dev.myst.ai/v1alpha2/openapi.json"
SENTINEL_DELETION_MARKER = "# DELETE ME!!"

# Note: The `openapi_python_client` library uses Open API tags to determine the sub-packages to create in the generated
#   endpoints module. It removes the `.` delimiter in tags but we want to create nested packages based on this
#   delimiter. This maps sanitized tags to their original tags so that this script can use the original tags to create
#   the appropriate nested packages. Note that new tags added to the Myst Platform Open API spec must be *manually*
#   added to this mapping.
PACKAGE_NAME_MAP: Dict[str, str] = {
    "aliases": "aliases",
    "connectors": "connectors",
    "organizations": "organizations",
    "projects": "projects",
    "projects_backtests": "projects.backtests",
    "projects_backtests_jobs": "projects.backtests.jobs",
    "projects_backtests_results": "projects.backtests.results",
    "projects_deployments": "projects.deployments",
    "projects_edges": "projects.edges",
    "projects_hpos": "projects.hpos",
    "projects_hpos_jobs": "projects.hpos.jobs",
    "projects_hpos_results": "projects.hpos.results",
    "projects_models": "projects.models",
    "projects_models_fit_jobs": "projects.models.fit_jobs",
    "projects_models_fit_policies": "projects.models.fit_policies",
    "projects_models_fit_results": "projects.models.fit_results",
    "projects_models_inputs": "projects.models.inputs",
    "projects_models_run_results": "projects.models.run_results",
    "projects_nodes": "projects.nodes",
    "projects_nodes_jobs": "projects.nodes.jobs",
    "projects_nodes_policies": "projects.nodes.policies",
    "projects_nodes_results": "projects.nodes.results",
    "projects_operations": "projects.operations",
    "projects_operations_inputs": "projects.operations.inputs",
    "projects_sources": "projects.sources",
    "projects_time_series": "projects.time_series",
    "projects_time_series_layers": "projects.time_series.layers",
    "projects_time_series_run_jobs": "projects.time_series.run_jobs",
    "projects_time_series_run_policies": "projects.time_series.run_policies",
    "projects_time_series_run_results": "projects.time_series.run_results",
    "users": "users",
}


def _move_directory(old_path: pathlib.Path, new_path: pathlib.Path) -> None:
    """Moves a directory, recursively creating any nested directories in new path."""
    # Create new.
    new_path.mkdir(parents=True, exist_ok=True)

    # Copy files from old to new.
    for f_ in old_path.glob("*"):
        shutil.copy(f_, new_path)

    # Remove old.
    shutil.rmtree(old_path)


def _remove_all_files_marked_for_deletion(base_path: pathlib.Path) -> None:
    for root, _, files in os.walk(base_path):
        for file in files:
            if file.endswith(".py"):
                full_path = pathlib.Path(root) / file
                with open(full_path) as f:
                    contents = f.read().strip()

                if contents == SENTINEL_DELETION_MARKER:
                    typer.echo(f"Deleting {file}")
                    full_path.unlink()


def _fix_imports(base_path: pathlib.Path, relative_models_package_name: str, absolute_models_package_name: str) -> None:
    relative_import_pattern = rf"from \.+{relative_models_package_name}"
    absolute_import_pattern = f"from {absolute_models_package_name}"

    typer.echo("Making all model imports absolute")
    typer.echo("Redirecting some imports to known handwritten models")

    for root, _, files in os.walk(base_path):
        for file in files:
            if file.endswith(".py"):
                full_path = pathlib.Path(root) / file
                with open(full_path) as f:
                    file_contents = f.read()

                file_contents = re.sub(
                    pattern=relative_import_pattern, repl=absolute_import_pattern, string=file_contents
                )

                # These auto-generated models have handwritten overrides.
                for model_name in _KNOWN_HANDWRITTEN_MODELS:
                    file_contents = file_contents.replace(
                        f"{absolute_import_pattern}.{model_name} import ", f"from myst.models.{model_name} import "
                    )

                with open(full_path, "w") as f:
                    f.write(file_contents)


def _rename_packages(base_path: pathlib.Path) -> None:
    for path in base_path.iterdir():
        if not path.is_dir():
            continue

        old_dirname = path.name
        new_dirname = PACKAGE_NAME_MAP.get(old_dirname)

        # Same name, so skip!
        if new_dirname == old_dirname:
            continue

        if new_dirname is None:
            raise KeyError(
                f"Package name: `{old_dirname}` does not have a corresponding entry in package name map. If "
                f"this is a new collection of endpoints, their tag must be added to the map so that the client "
                f"generator can rename their package appropriately."
            )

        namespaced_dirname: str = new_dirname.replace(".", "/")

        typer.echo(f"Moving {old_dirname} -> {namespaced_dirname}")

        old_package_path = base_path / old_dirname
        new_package_path = base_path / namespaced_dirname

        _move_directory(old_package_path, new_package_path)


app = typer.Typer()


@app.command()
def main(openapi_url: str = OPENAPI_URL) -> None:
    """Generates the OpenAPI client library."""
    lib_path = PROJECT_ROOT / "myst"
    project_name = "openapi"
    api_dirname = "api"
    models_dirname = "models"

    # Create a (clean) place for the project to go.
    gen_client_container_path = lib_path / project_name
    typer.echo(f"Removing {gen_client_container_path}")
    shutil.rmtree(gen_client_container_path.absolute(), ignore_errors=True)

    # Generate a project config under our custom name.
    config = Config()
    config.project_name_override = project_name

    # Custom Jinja2 templates live here.
    custom_template_dir = TOOLS_ROOT / "client-gen-templates"

    # Generate the openapi client in the correct directory.
    os.chdir(lib_path.absolute())
    typer.echo(f"Creating new client in {gen_client_container_path} based on OpenAPI schema at {openapi_url}")
    create_new_client(
        url=openapi_url, path=None, meta=MetaType.NONE, config=config, custom_template_path=custom_template_dir
    )

    typer.echo(f'Removing all files with sentinel deletion marker "{SENTINEL_DELETION_MARKER}"')
    _remove_all_files_marked_for_deletion(gen_client_container_path)

    typer.echo(f"Fixing imports")

    _fix_imports(
        gen_client_container_path,
        relative_models_package_name=models_dirname,
        absolute_models_package_name=f"{lib_path.name}.{project_name}.{models_dirname}",
    )

    typer.echo(f"Renaming default {api_dirname} package names")

    _rename_packages(gen_client_container_path / api_dirname)

    typer.echo("All done.")


if __name__ == "__main__":
    app()



from typing import List, Optional

from myst.models import base_model


class TimeDatasetRow(base_model.BaseModel):
    """"""

    start_time: str
    end_time: str
    as_of_time: str
    values: List[float]
    mask: Optional[List[float]] = None



from myst.client import Client
from myst.openapi.models.operation_get import OperationGet
from myst.openapi.models.operation_update import OperationUpdate


def request_sync(client: Client, project_uuid: str, uuid: str, json_body: OperationUpdate) -> OperationGet:
    """Updates an operation."""

    return client.request(
        method="patch",
        path=f"/projects/{project_uuid}/operations/{uuid}",
        response_class=OperationGet,
        request_model=json_body,
    )



import json
from typing import Any, Dict, Optional

import httpx
import pandas as pd
from pydantic import BaseModel

from myst.resources.utils import backtest_result_to_pandas_data_frame


def _download_result(result_url: str) -> Dict[str, Any]:
    """Downloads the backtest result from the supplied URL and parses it."""
    response = httpx.get(result_url)

    if response.status_code == 200:
        result_data = json.loads(response.content)
    else:
        raise RuntimeError("Could not download backtest result.")

    return result_data


class HPOTrial(BaseModel):
    parameters: Dict[str, Any]
    metrics: Dict[str, float]
    create_time: Optional[str] = None
    backtest_result_url: Optional[str] = None

    # TODO(CRYS-3649): move `backtest_result_to_pandas_data_frame` logic out of client.
    def to_pandas_data_frame(self) -> pd.DataFrame:
        """Downloads the HPO result and converts the time arrays to pandas data frames.

        Data will be re-indexed against the predictions' natural time index, dropping any target data that doesn't
        correspond to a prediction.

        Returns:
            a pandas data frame with the predictions made by the HPO, and their corresponding targets. # noqa: DAR202

        Raises:
            ValueError: if missing `backtest_result_url`
        """
        if self.backtest_result_url is None:
            raise ValueError("No result url saved for this `HPOTrial`.")

        result_data = _download_result(result_url=self.backtest_result_url)
        pandas_data_frame = backtest_result_to_pandas_data_frame(result_data=result_data)

        return pandas_data_frame



from pydantic import Field
from typing_extensions import Literal

from myst.models import base_model


class CronTimingGet(base_model.BaseModel):
    """Cron timing schema for get responses."""

    object_: Literal["Timing"] = Field(..., alias="object")
    type: Literal["CronTiming"]
    cron_expression: str
    time_zone: str



from typing import Optional

from pydantic import Field
from typing_extensions import Literal

from myst.models import base_model


class OrganizationUpdate(base_model.BaseModel):
    """Schema for organization update requests."""

    object_: Optional[Literal["Organization"]] = Field(..., alias="object")
    uuid: Optional[str] = None
    create_time: Optional[str] = None
    update_time: Optional[str] = None
    name: Optional[str] = None



from myst.client import Client
from myst.openapi.models.model_fit_policy_list import ModelFitPolicyList


def request_sync(client: Client, project_uuid: str, model_uuid: str) -> ModelFitPolicyList:
    """Lists fit policies for a model."""

    return client.request(
        method="get",
        path=f"/projects/{project_uuid}/models/{model_uuid}/fit_policies/",
        response_class=ModelFitPolicyList,
    )



from myst.client import Client
from myst.openapi.models.project_get import ProjectGet


def request_sync(client: Client, uuid: str) -> ProjectGet:
    """Gets a project by its unique identifier."""

    return client.request(method="get", path=f"/projects/{uuid}", response_class=ProjectGet)



from typing import List, Union

from myst.models import base_model
from myst.openapi.models.model_connector_get import ModelConnectorGet
from myst.openapi.models.operation_connector_get import OperationConnectorGet
from myst.openapi.models.source_connector_get import SourceConnectorGet


class PolymorphicConnectorList(base_model.BaseModel):
    """Schema for polymorphic connector list responses."""

    data: List[Union[SourceConnectorGet, OperationConnectorGet, ModelConnectorGet]]
    has_more: bool



from typing import Optional

from pydantic import Field
from typing_extensions import Literal

from myst.models import base_model
from myst.openapi.models.deploy_status import DeployStatus
from myst.openapi.models.org_sharing_role import OrgSharingRole


class ProjectUpdate(base_model.BaseModel):
    """Schema for project update requests."""

    object_: Optional[Literal["Project"]] = Field(..., alias="object")
    uuid: Optional[str] = None
    create_time: Optional[str] = None
    update_time: Optional[str] = None
    organization: Optional[str] = None
    owner: Optional[str] = None
    title: Optional[str] = None
    description: Optional[str] = None
    organization_sharing_enabled: Optional[bool] = None
    organization_sharing_role: Optional[OrgSharingRole] = None
    creator: Optional[str] = None
    deploy_status: Optional[DeployStatus] = None



from myst.client import Client
from myst.openapi.models.layer_get import LayerGet
from myst.openapi.models.layer_update import LayerUpdate


def request_sync(
    client: Client, project_uuid: str, time_series_uuid: str, uuid: str, json_body: LayerUpdate
) -> LayerGet:
    """Updates an existing layer for a time series."""

    return client.request(
        method="patch",
        path=f"/projects/{project_uuid}/time_series/{time_series_uuid}/layers/{uuid}",
        response_class=LayerGet,
        request_model=json_body,
    )



import time
from typing import TYPE_CHECKING, List, Optional, Union
from uuid import UUID

from myst import get_client
from myst.adapters.utils import get_resource_uuid
from myst.models.types import UUIDOrStr
from myst.openapi.api.projects.models.fit_jobs import get_models_fit_job
from myst.openapi.api.projects.time_series.run_jobs import get_time_series_run_job
from myst.openapi.models.job_error import JobError
from myst.openapi.models.job_state import JobState
from myst.resources.resource import Resource

if TYPE_CHECKING:  # Avoid circular imports.
    from myst.resources.model import Model
    from myst.resources.project import Project
    from myst.resources.time_series import TimeSeries


class NodeJob(Resource):
    """Describes a job associated with a node.

    Attributes:
        project: the UUID of the project this job corresponds to
        node: the UUID of the node this job corresponds to
        start_timing: the beginning of the natural time range for which this job applies, inclusive
        end_timing: the end of the natural time range for which this job applies, exclusive
        as_of_time: the as of time of this job applies
        result: the UUID of the optional result this job corresponds to
        errors: a list of errors associated with this job
        state: job state
    """

    project: UUID
    node: UUID
    start_time: str
    end_time: str
    as_of_time: str
    result: Optional[UUID]
    errors: List[JobError]
    state: JobState

    def is_completed(self) -> bool:
        """Returns `True` if the job has completed.

        Completed jobs include both successful and failed jobs.

        Returns:
            whether the job has completed
        """
        return self.state in [JobState.SUCCESS, JobState.FAILURE, JobState.UPSTREAM_FAILURE]


class TimeSeriesRunJob(NodeJob):
    """Describes a node run job on a time series."""

    @classmethod
    def get(
        cls, project: Union["Project", UUIDOrStr], time_series: Union["TimeSeries", UUIDOrStr], uuid: UUIDOrStr
    ) -> "TimeSeriesRunJob":
        """Gets a specific time series run job by its identifier."""
        time_series_job = get_time_series_run_job.request_sync(
            client=get_client(),
            project_uuid=str(get_resource_uuid(project)),
            time_series_uuid=str(get_resource_uuid(time_series)),
            uuid=str(uuid),
        )

        return TimeSeriesRunJob.parse_obj(dict(time_series_job.dict(), project=get_resource_uuid(project)))

    def wait_until_completed(self) -> "TimeSeriesRunJob":
        """Returns the job once it has completed."""
        while True:
            refreshed_job = TimeSeriesRunJob.get(project=self.project, time_series=self.node, uuid=self.uuid)

            if refreshed_job.is_completed():
                break

            time.sleep(60)

        return TimeSeriesRunJob.parse_obj(dict(refreshed_job.dict(), project=self.project))


class ModelFitJob(NodeJob):
    """Describes a model fit job on a model."""

    @classmethod
    def get(
        cls, project: Union["Project", UUIDOrStr], model: Union["Model", UUIDOrStr], uuid: UUIDOrStr
    ) -> "ModelFitJob":
        """Gets a specific model fit job by its identifier."""
        model_job = get_models_fit_job.request_sync(
            client=get_client(),
            project_uuid=str(get_resource_uuid(project)),
            model_uuid=str(get_resource_uuid(model)),
            uuid=str(uuid),
        )

        return ModelFitJob.parse_obj(dict(model_job.dict(), project=get_resource_uuid(project)))

    def wait_until_completed(self) -> "ModelFitJob":
        """Returns the job once it has completed."""
        while True:
            refreshed_job = ModelFitJob.get(project=self.project, model=self.node, uuid=self.uuid)

            if refreshed_job.is_completed():
                break

            time.sleep(60)

        return ModelFitJob.parse_obj(dict(refreshed_job.dict(), project=self.project))



import enum
from typing import List
from uuid import UUID

from myst.connectors.source_connector import SourceConnector

CLEANED_OBSERVATIONS_ID = UUID("cfeb3606-217d-4fd0-9abf-21cd04a8e9e7")


@enum.unique
class Field(str, enum.Enum):
    APPARENT_TEMPERATURE_CELSIUS = "apparentTemperatureCelsius"
    CLOUD_COVERAGE_PERCENT = "cloudCoveragePercent"
    SURFACE_DEWPOINT_TEMPERATURE_CELSIUS = "surfaceDewpointTemperatureCelsius"
    DIFFUSE_HORIZONTAL_RADIATION_WSQM = "diffuseHorizontalRadiationWsqm"
    DIRECT_NORMAL_IRRADIANCE_WSQM = "directNormalIrradianceWsqm"
    DOWNWARD_SOLAR_RADIATION_WSQM = "downwardSolarRadiationWsqm"
    HEAT_INDEX_CELSIUS = "heatIndexCelsius"
    MSL_PRESSURE_KILOPASCALS = "mslPressureKilopascals"
    PRECIPITATION_PREVIOUS_HOUR_CENTIMETERS = "precipitationPreviousHourCentimeters"
    RELATIVE_HUMIDITY_PERCENT = "relativeHumidityPercent"
    SNOWFALL_CENTIMETERS = "snowfallCentimeters"
    SURFACE_AIR_PRESSURE_KILOPASCALS = "surfaceAirPressureKilopascals"
    SURFACE_TEMPERATURE_CELSIUS = "surfaceTemperatureCelsius"
    SURFACE_WIND_GUSTS_KPH = "surfaceWindGustsKph"
    SURFACE_WET_BULB_TEMPERATURE_CELSIUS = "surfaceWetBulbTemperatureCelsius"
    WIND_CHILL_TEMPERATURE_CELSIUS = "windChillTemperatureCelsius"
    WIND_DIRECTION_DEGREES = "windDirectionDegrees"
    WIND_SPEED_KPH = "windSpeedKph"


class CleanedObservations(SourceConnector):
    def __init__(self, latitude: float, longitude: float, fields: List[Field]) -> None:
        super().__init__(
            uuid=CLEANED_OBSERVATIONS_ID, parameters=dict(latitude=latitude, longitude=longitude, fields=fields)
        )



from typing import TYPE_CHECKING, Any, Dict, List, Optional, Union

from myst.adapters.timing import to_timing_create
from myst.adapters.utils import drop_unset_values, get_resource_uuid
from myst.client import get_client
from myst.core.data.time_array import TimeArray
from myst.core.time.time import Time
from myst.core.time.time_delta import TimeDelta
from myst.models.time_dataset import TimeDataset, TimeDatasetRow
from myst.models.timing import ScheduleTiming, TimeRangeBoundary, from_time_range_boundary
from myst.models.types import (
    UNSET,
    AxisLabels,
    CoordinateLabels,
    ItemOrSlice,
    OptionalArgument,
    Shape,
    UUIDOrStr,
    to_uuid,
)
from myst.openapi.api.projects.time_series import (
    create_time_series,
    get_time_series,
    insert_time_series_data,
    list_all_time_series,
    query_time_series_data,
    time_series_run_job_create,
    update_time_series,
)
from myst.openapi.models.node_job_create import NodeJobCreate
from myst.openapi.models.time_series_create import TimeSeriesCreate
from myst.openapi.models.time_series_insert import TimeSeriesInsert
from myst.openapi.models.time_series_update import TimeSeriesUpdate
from myst.resources.job import TimeSeriesRunJob
from myst.resources.layer import TimeSeriesLayer
from myst.resources.node import Node
from myst.resources.policy import TimeSeriesRunPolicy
from myst.resources.result import TimeSeriesRunResult, TimeSeriesRunResultMetadata

if TYPE_CHECKING:  # Avoid circular imports.
    from myst.recipes.time_series_recipe import TimeSeriesRecipe
    from myst.resources.project import Project


class TimeSeries(Node):
    """A queryable, shareable node containing fixed-frequency, arbitrarily dimensioned time series data.

    A time series stitches together data from multiple layers into a single stream, by consuming data from lower-
    precedence layers when data from higher-precedence layers is not available, and by consuming only the upstream data
    that is within the time range bounds set on each layer.

    In addition to queries, a time series also supports inserts, which take precedence over data flowing through the
    input layers.

    Attributes:
        sample_period: the frequency at which the data is sampled
        cell_shape: the shape of the data in each cell
        coordinate_labels: the labels for each coordinate of a cell
        axis_labels: the labels for each axis of a cell
    """

    sample_period: TimeDelta
    cell_shape: Shape
    coordinate_labels: CoordinateLabels
    axis_labels: AxisLabels

    @classmethod
    def create(
        cls,
        project: Union["Project", UUIDOrStr],
        title: str,
        sample_period: TimeDelta,
        cell_shape: Shape = (),
        coordinate_labels: CoordinateLabels = (),
        axis_labels: AxisLabels = (),
        description: Optional[str] = None,
    ) -> "TimeSeries":
        """Creates a new time series node.

        Args:
            project: the project in which to create the time series
            title: the title of the time series
            sample_period: the frequency at which the data is sampled
            cell_shape: the shape of the data in each cell of the time series
            coordinate_labels: the labels for each coordinate of a cell in the time series
            axis_labels: the labels for each axis of a cell in the time series
            description: a brief description of the time series

        Returns:
            the newly created time series
        """
        time_series = create_time_series.request_sync(
            client=get_client(),
            project_uuid=str(get_resource_uuid(project)),
            json_body=TimeSeriesCreate(
                object="Node",
                type="TimeSeries",
                title=title,
                sample_period=sample_period.to_iso_string(),
                cell_shape=cell_shape,
                coordinate_labels=coordinate_labels,
                axis_labels=axis_labels,
                description=description,
            ),
        )

        return TimeSeries.parse_obj(time_series.dict())

    @classmethod
    def create_from_recipe(
        cls,
        project: Union["Project", UUIDOrStr],
        recipe: "TimeSeriesRecipe",
        title: Optional[str] = None,
        description: Optional[str] = None,
    ) -> "TimeSeries":
        """Creates a new time series node based on a provided recipe.

        Args:
            project: the project in which to create the time series
            recipe: the recipe used to create the time series
            title: the title of the time series; if none is given, the recipe will supply a sensible default
            description: the description of the time series

        Returns:
            the newly created time series
        """
        return recipe.create(project=project, title=title, description=description)

    @classmethod
    def get(cls, project: Union["Project", UUIDOrStr], uuid: UUIDOrStr) -> "TimeSeries":
        """Gets a specific time series by its identifier."""
        time_series = get_time_series.request_sync(
            client=get_client(), project_uuid=str(get_resource_uuid(project)), uuid=str(to_uuid(uuid))
        )

        return TimeSeries.parse_obj(time_series.dict())

    def insert_time_array(self, time_array: TimeArray) -> None:
        """Inserts data from a `TimeArray` into this time series.

        The data from this manual insertion will take precedence over the other layers flowing into this time series.

        Args:
            time_array: the array of data to be inserted
        """
        time_dataset_row = TimeDatasetRow(
            start_time=time_array.start_time,
            end_time=time_array.end_time,
            as_of_time=time_array.as_of_time,
            values=time_array.values,
            mask=time_array.mask,
        )

        time_dataset = TimeDataset(
            object="TimeDataset",
            cell_shape=time_array.cell_shape,
            sample_period=time_array.sample_period,
            data=[time_dataset_row],
            coordinate_labels=time_array.coordinate_labels,
            axis_labels=time_array.axis_labels,
        )

        insert_time_series_data.request_sync(
            client=get_client(),
            project_uuid=str(self.project),
            uuid=str(self.uuid),
            json_body=TimeSeriesInsert(time_dataset=time_dataset),
        )

    @classmethod
    def list(cls, project: Union["Project", UUIDOrStr]) -> List["TimeSeries"]:
        """Gets all time series by project."""
        all_time_series = list_all_time_series.request_sync(client=get_client())

        return [
            TimeSeries.parse_obj(time_series.dict())
            for time_series in all_time_series.data
            if str(time_series.project) == str(get_resource_uuid(project))
        ]

    def update(
        self,
        title: OptionalArgument[str] = UNSET,
        sample_period: OptionalArgument[TimeDelta] = UNSET,
        cell_shape: OptionalArgument[Shape] = UNSET,
        coordinate_labels: OptionalArgument[CoordinateLabels] = UNSET,
        axis_labels: OptionalArgument[AxisLabels] = UNSET,
        description: OptionalArgument[str] = UNSET,
    ) -> "TimeSeries":
        """Updates time series."""
        time_series_update: Dict[str, Any] = dict(
            object="Node",
            type="TimeSeries",
            title=title,
            sample_period=sample_period.to_iso_string() if isinstance(sample_period, TimeDelta) else sample_period,
            cell_shape=cell_shape,
            coordinate_labels=coordinate_labels,
            axis_labels=axis_labels,
            description=description,
        )

        time_series = update_time_series.request_sync(
            client=get_client(),
            project_uuid=str(self.project),
            uuid=str(self.uuid),
            json_body=TimeSeriesUpdate.parse_obj(drop_unset_values(time_series_update)),
        )

        return TimeSeries.parse_obj(time_series.dict())

    def query_time_array(
        self,
        start_time: Time,
        end_time: Time,
        as_of_time: Optional[Time] = None,
        as_of_offset: Optional[TimeDelta] = None,
    ) -> TimeArray:
        """Queries this time series for data according to the given parameters.

        At most one of `as_of_time` and `as_of_offset` may be specified. If `as_of_time` is specified, the query will
        return data as of that particular time. If `as_of_offset` is specified, the query will return data for a
        constant offset between the "natural" time and the as of time. If neither is specified, the API assumes an
        `as_of_time` of now.

        Args:
            start_time: the beginning of the natural time range to query over, inclusive
            end_time: the end of the natural time range to query over, exclusive
            as_of_time: the precise as of time to query
            as_of_offset: the offset from as of time to query by, where as_of_offset = natural_time - as_of_time

        Returns:
            a time array containing data for the specified time range and as of time

        Raises:
            ValueError: if both `as_of_time` and `as_of_offset` are specified
        """
        if as_of_time and as_of_offset:
            raise ValueError("At most one of `as_of_time` and `as_of_offset` may be specified.")

        time_series_query_result_get = query_time_series_data.request_sync(
            client=get_client(),
            project_uuid=str(self.project),
            uuid=str(self.uuid),
            start_time=start_time.to_iso_string(),
            end_time=end_time.to_iso_string(),
            as_of_time=as_of_time.to_iso_string() if as_of_time is not None else None,
            as_of_offset=as_of_offset.to_iso_string() if as_of_offset is not None else None,
        )

        time_dataset = TimeDataset.parse_obj(time_series_query_result_get.time_dataset)

        return time_dataset.flatten()

    def create_layer(
        self,
        node: Union[Node, UUIDOrStr],
        order: int,
        output_index: int = 0,
        label_indexer: Optional[ItemOrSlice] = None,
        start_timing: Optional[TimeRangeBoundary] = None,
        end_timing: Optional[TimeRangeBoundary] = None,
    ) -> TimeSeriesLayer:
        """Creates a layer into this time series.

        Args:
            node: the node whose data will flow into this time series
            order: integer specifying priority of this layer when combining multiple layers; lower order implies
                higher precedence
            output_index: which time dataset, out of the sequence of upstream time datasets, to pass to this time
                series
            label_indexer: the slice of the upstream data to pass to this time series
            start_timing: the beginning of the natural time range this layer should produce data for; if None, there
                is no restriction on the beginning of the range
            end_timing: the end of the natural time range this layer should produce data for; if None, there is no
                restriction on the end of the range

        Returns:
            the newly created time series layer
        """
        return TimeSeriesLayer.create(
            project=self.project,
            time_series=self.uuid,
            node=node,
            order=order,
            output_index=output_index,
            label_indexer=label_indexer,
            start_timing=start_timing,
            end_timing=end_timing,
        )

    def list_layers(self) -> List[TimeSeriesLayer]:
        """Lists all layers into this time series."""
        return TimeSeriesLayer.list(project=self.project, time_series=self)

    def create_run_policy(
        self,
        schedule_timing: Union[Time, TimeDelta, ScheduleTiming],
        start_timing: TimeRangeBoundary,
        end_timing: TimeRangeBoundary,
        active: bool = True,
    ) -> TimeSeriesRunPolicy:
        return TimeSeriesRunPolicy.create(
            project=self.project,
            time_series=self.uuid,
            schedule_timing=schedule_timing,
            start_timing=start_timing,
            end_timing=end_timing,
            active=active,
        )

    def list_run_policies(self) -> List[TimeSeriesRunPolicy]:
        return TimeSeriesRunPolicy.list(project=self.project, time_series=self.uuid)

    def list_run_results(self) -> List[TimeSeriesRunResultMetadata]:
        return TimeSeriesRunResult.list(project=self.project, time_series=self)

    def get_run_result(self, uuid: UUIDOrStr) -> TimeSeriesRunResult:
        return TimeSeriesRunResult.get(project=self.project, time_series=self, uuid=uuid)

    def run(self, start_timing: TimeRangeBoundary, end_timing: TimeRangeBoundary) -> TimeSeriesRunJob:
        """Create an ad hoc node run job for this time series."""
        node_run_job = time_series_run_job_create.request_sync(
            client=get_client(),
            project_uuid=str(self.project),
            uuid=str(self.uuid),
            json_body=NodeJobCreate(
                object="NodeJob",
                start_timing=to_timing_create(from_time_range_boundary(start_timing)),
                end_timing=to_timing_create(from_time_range_boundary(end_timing)),
            ),
        )

        return TimeSeriesRunJob.parse_obj(dict(node_run_job.dict(), project=self.project))



from typing import List

from myst.models import base_model
from myst.openapi.models.time_series_get import TimeSeriesGet


class TimeSeriesList(base_model.BaseModel):
    """Schema for time series list responses."""

    data: List[TimeSeriesGet]
    has_more: bool



"""This modules contains custom Myst exceptions."""


class BaseMystError(Exception):
    """Base class for any exceptions raised in this library."""


class MystClientError(BaseMystError):
    """Generic errors in the Myst client."""


class UnauthenticatedError(MystClientError):
    """Raised when authentication is missing or fails."""

    def __init__(self, message: str) -> None:
        """An `UnauthenticatedError` is initialized with a message."""
        super().__init__(message)
        self.message = message


class NotFound(MystClientError):
    """The requested resource was not found."""

    def __init__(self, message: str) -> None:
        super().__init__(message)
        self.message = message


class MystAPIError(BaseMystError):
    """Generic Myst API error."""

    def __init__(self, status_code: int, message: str) -> None:
        """A `MystAPIError` is initialized with an HTTP status code and a message."""
        super().__init__(message)
        self.status_code = status_code
        self.message = message



from typing import Optional

from pydantic import Field
from typing_extensions import Literal

from myst.models import base_model
from myst.openapi.models.organization_role import OrganizationRole


class UserUpdate(base_model.BaseModel):
    """Schema for user update requests."""

    object_: Optional[Literal["User"]] = Field(..., alias="object")
    uuid: Optional[str] = None
    create_time: Optional[str] = None
    update_time: Optional[str] = None
    email: Optional[str] = None
    organization: Optional[str] = None
    organization_role: Optional[OrganizationRole] = None



from typing import List

from myst.models import base_model
from myst.openapi.models.model_fit_policy_get import ModelFitPolicyGet


class ModelFitPolicyList(base_model.BaseModel):
    """Schema for model fit policy list responses."""

    data: List[ModelFitPolicyGet]
    has_more: bool



from typing import Any, List, Optional, Union

from pydantic import Field
from typing_extensions import Literal

from myst.models import base_model


class InputGet(base_model.BaseModel):
    """Schema for input get responses."""

    object_: Literal["Edge"] = Field(..., alias="object")
    uuid: str
    create_time: str
    type: Literal["Input"]
    downstream_node: str
    upstream_node: str
    output_index: int
    group_name: str
    update_time: Optional[str] = None
    label_indexer: Optional[Union[List[Union[int, str, List[Any]]], List[Any], int, str]] = None



from myst.client import Client
from myst.openapi.models.project_list import ProjectList


def request_sync(client: Client) -> ProjectList:
    """Lists projects."""

    return client.request(method="get", path=f"/projects/", response_class=ProjectList)



from typing import Any, List, Optional, Union

from pydantic import Field
from typing_extensions import Literal

from myst.models import base_model
from myst.openapi.models.absolute_timing_create import AbsoluteTimingCreate
from myst.openapi.models.relative_timing_create import RelativeTimingCreate


class LayerCreate(base_model.BaseModel):
    """Schema for layer create requests."""

    upstream_node: str
    order: int
    object_: Optional[Literal["Edge"]] = Field("Edge", alias="object")
    type: Optional[Literal["Layer"]] = "Layer"
    output_index: Optional[int] = 0
    label_indexer: Optional[Union[List[Union[int, str, List[Any]]], List[Any], int, str]] = None
    start_timing: Optional[Union[AbsoluteTimingCreate, RelativeTimingCreate]] = None
    end_timing: Optional[Union[AbsoluteTimingCreate, RelativeTimingCreate]] = None



import enum
from uuid import UUID

from myst.connectors.operation_connector import OperationConnector


@enum.unique
class GroupName(str, enum.Enum):
    LOC = "loc"
    SCALE = "scale"


@enum.unique
class Function(str, enum.Enum):
    CDF = "CDF"
    PDF = "PDF"
    PPF = "PPF"


class NormalDistribution(OperationConnector):
    def __init__(self, function: Function, value: float) -> None:
        super().__init__(
            uuid=UUID("b5bee009-db1c-448a-b701-e0345a96db6b"), parameters=dict(function=function, value=value)
        )



from typing import Optional

from pydantic import Field
from typing_extensions import Literal

from myst.models import base_model
from myst.openapi.models.operation_connector_get_parameters_schema import OperationConnectorGetParametersSchema


class OperationConnectorGet(base_model.BaseModel):
    """Schema for operation connector get responses."""

    object_: Literal["Connector"] = Field(..., alias="object")
    type: Literal["OperationConnector"]
    uuid: str
    title: str
    provider: str
    description: str
    parameters_schema: OperationConnectorGetParametersSchema
    icon_url: Optional[str] = None



from myst.client import Client
from myst.openapi.models.model_run_result_get import ModelRunResultGet


def request_sync(client: Client, project_uuid: str, model_uuid: str, uuid: str) -> ModelRunResultGet:
    """Gets a model run result by its unique identifier."""

    return client.request(
        method="get",
        path=f"/projects/{project_uuid}/models/{model_uuid}/run_results/{uuid}",
        response_class=ModelRunResultGet,
    )



from typing import Optional, Union

from pydantic import Field
from typing_extensions import Literal

from myst.models import base_model
from myst.openapi.models.absolute_timing_create import AbsoluteTimingCreate
from myst.openapi.models.cron_timing_create import CronTimingCreate
from myst.openapi.models.hyperopt_create import HyperoptCreate
from myst.openapi.models.relative_timing_create import RelativeTimingCreate
from myst.openapi.models.search_space import SearchSpace


class HPOCreate(base_model.BaseModel):
    """HPO schema for create input."""

    object_: Literal["HPO"] = Field(..., alias="object")
    title: str
    model: str
    search_algorithm: HyperoptCreate
    fit_start_timing: Union[AbsoluteTimingCreate, RelativeTimingCreate]
    fit_end_timing: Union[AbsoluteTimingCreate, RelativeTimingCreate]
    predict_start_timing: Union[AbsoluteTimingCreate, RelativeTimingCreate]
    predict_end_timing: Union[AbsoluteTimingCreate, RelativeTimingCreate]
    description: Optional[str] = None
    search_space: Optional[SearchSpace] = None
    test_start_time: Optional[str] = None
    test_end_time: Optional[str] = None
    fit_reference_timing: Optional[Union[AbsoluteTimingCreate, CronTimingCreate]] = None
    predict_reference_timing: Optional[CronTimingCreate] = None



from typing import Optional

from pydantic import Field
from typing_extensions import Literal

from myst.models import base_model
from myst.openapi.models.deploy_status import DeployStatus
from myst.openapi.models.org_sharing_role import OrgSharingRole


class ProjectGet(base_model.BaseModel):
    """Schema for project get responses."""

    object_: Literal["Project"] = Field(..., alias="object")
    uuid: str
    create_time: str
    organization: str
    owner: str
    title: str
    creator: str
    deploy_status: DeployStatus
    organization_sharing_enabled: bool
    update_time: Optional[str] = None
    description: Optional[str] = None
    organization_sharing_role: Optional[OrgSharingRole] = None



from myst.client import Client
from myst.openapi.models.layer_create import LayerCreate
from myst.openapi.models.layer_get import LayerGet


def request_sync(client: Client, project_uuid: str, time_series_uuid: str, json_body: LayerCreate) -> LayerGet:
    """Creates a layer for a time series."""

    return client.request(
        method="post",
        path=f"/projects/{project_uuid}/time_series/{time_series_uuid}/layers/",
        response_class=LayerGet,
        request_model=json_body,
    )



from myst.models import base_model
from myst.openapi.models.alias_get_type import AliasGetType


class AliasGet(base_model.BaseModel):
    """Schema for Alias get responses."""

    alias: str
    type: AliasGetType
    subject: str
